import os
import sys
import glob
import multiprocessing
import datetime
from astropy.io import fits
from collections import defaultdict
from copy import deepcopy
import refstis
from stistools import basic2d, calstis
import archive_dark_query
import StisPixCteCorr
from crds.bestrefs import BestrefsScript

__author__  = 'Sean Lockwood'
__version__ = '1.1'

crds_server_url = 'https://hst-crds.stsci.edu'

class FileError(Exception):
    pass

class VersionError(Exception):
    pass

def stis_cti(science_dir, dark_dir, ref_dir, num_processes, pctetab=None, 
             all_weeks_flag=False, allow=False, clean=False, clean_all=False, 
             crds_update=False, ignore_missing=False, verbose=1):
    '''
    Runs the HST/STIS/CCD pixel-based CTI-correction on science data and 
    component darks, generating and applying a CTI-corrected super-dark in 
    the process.
    
    Documentation is available at http://pythonhosted.org/stis_cti/
    
    :param science_dir:
        Directory containing uncalibrated science data to be corrected.
    :param dark_dir: str
        Directory containing calibrated component darks to be corrected and
        location where CTI-corrected component darks are placed.
    :param ref_dir: str
        Directory where CTI-corrected super-darks are placed.  These will be
        used again unless they are deleted or clean_all=True.
    :param num_processes:
        Max number of parallel processes to use when running the CTI-correction 
        algorithm.
    :param pctetab:
        The path + name of the PCTETAB reference file to use in the CTI-correction.  
        If not specified, one is selected from (1) the ref_dir, or (2) from the 
        package data directory.  The last file (alphabetically) is chosen.
    :param all_weeks_flag:
        UNTESTED.  Generates weekdarks for all weeks within each annealing period 
        to be processed.
    :param allow:
        UNTESTED.  Use more lenient filtering when determining which files should 
        be allowed to be corrected.
    :param clean:
        Remove intermediate and final products in the science_dir from previous 
        runs of this script.
    :param clean_all:
        'clean' + remove CTI-corrected super-darks and component darks before 
        reprocessing.
    :param crds_update:
        Runs crds.bestrefs script to update science file headers and download 
        pipeline reference files.
    :param ignore_missing:
        Ignore missing dark files.  This is useful for when annealing month contains
        amp=A RAW files, but no associated FLT files.
    :param verbose:
        Verbosity of text printed to the screen and saved in the log file.
    
    :type science_dir: str
    :type dark_dir: str
    :type ref_dir: str
    :type num_processes: int
    :type pctetab: str, optional 
    :type all_weeks_flag: bool, optional
    :type allow: bool, optional
    :type clean: bool, optional
    :type clean_all: bool, optional
    :type crds_update: bool, optional
    :type ignore_missing: bool, optional, default=False
    :type verbose: int {0,1,2}, optional, default=1
    
    .. note::
      Unless crds_update is True, the $oref shell variable must be set to the directory of 
      STIS standard pipeline reference files.
      
    .. note::
      Note that the 'all_week_flag' and 'allow' options have not been tested!
    '''
    # Open a log file and copy {STDOUT, STDERR} to it:
    log = Logger(os.path.join(science_dir, 'cti_{}.log'.format(datetime.datetime.now().isoformat('_'))))
    
    try:
        import pkg_resources
    except ImportError:
        pass
    try:
        from platform import node, platform
        sys_info = '{}; {}'.format(node(), platform())
    except ImportError:
        sys_info = 'Indeterminate'
    
    start_time = datetime.datetime.now()
    
    # Print system information:
    print 'Running CTI-correction script:  {} v{}'.format(os.path.basename(__file__), __version__)
    print 'System:                         {}'.format(sys_info)
    print 'Number of parallel processes:   {}'.format(num_processes)
    if clean:
        print '--clean'
    if clean_all:
        print '--clean_all'
    if crds_update:
        print '--crds_update'
    if ignore_missing:
        print '--ignore_missing'
    if verbose:
        print 'verbose mode:                   {}'.format(verbose)

    print 'Start time:                     {}\n'.format(start_time.isoformat(' '))
    
    # Check that directories exist:
    if not os.path.isdir(science_dir):
        raise FileError('science_dir does not exist:  ' + science_dir)
    if not os.path.isdir(dark_dir):
        raise FileError('dark_dir does not exist:  '    + dark_dir)
    if not os.path.isdir(ref_dir):
        raise FileError('ref_dir does not exist:  '     + ref_dir)
    
    if clean_all:
        clean = True
    
    # Check PCTETAB:
    if pctetab is None:
        pctetabs = glob.glob(os.path.join(ref_dir, '*_pcte.fits*'))
        pctetabs.sort()
        if len(pctetabs) > 0:
            pctetab = pctetabs[-1]
        else:
            # Couldn't find PCTETAB in ref/ directory, so use the PCTETAB included with the stis_cti package:
            try:
                default_pcte_str = pkg_resources.resource_filename(stis_cti.__name__, 'data/*_pcte.fits')
            except NameError:
                default_pcte_str = os.path.join(os.path.dirname(stis_cti.__file__), 'data', '*_pcte.fits')
            package_pctes = glob.glob(default_pcte_str)
            if len(package_pctes) == 0:
                raise FileError('Couldn\'t find a PCTETAB in {}\nor package default in {}'.format(
                    ref_dir, os.path.dirname(default_pcte_str)))
            package_pctes.sort()
            pctetab = package_pctes[-1]
            print 'WARNING:  Using package-default PCTETAB:\n{}\n'.format(pctetab)
    if not os.path.exists(pctetab):
        raise FileError('PCTETAB does not exist:  ' + pctetab)
    
    if verbose:
        print 'cwd         = {}'.format(os.getcwd() + os.path.sep)
        print 'science_dir = {}'.format(science_dir)
        print 'dark_dir    = {}'.format(dark_dir)
        print 'ref_dir     = {}'.format(ref_dir)
        print 'PCTETAB     = {}\n'.format(pctetab)
    log.flush()
    
    # Put dark_dir and PCTETAB's dir in environmental variables:
    if '$' not in ref_dir:
        os.environ['ctirefs'] = os.path.abspath(ref_dir) + os.path.sep
        ref_dir = '$ctirefs' + os.path.sep
    if '$' not in pctetab:
        os.environ['pctetab'] = os.path.abspath(os.path.dirname(pctetab)) + os.path.sep
        # dir$file format needed for StisPixCteCorr()...
        pctetab = 'pctetab$' + os.path.basename(pctetab)  # This won't work with os.path.expandvars()!
    
    # Check PCTETAB's version against min/max allowed by this code:
    if not allow:
        check_pctetab_version(pctetab, verbose)
    
    # Setup environmental variables for CRDS code to update headers:
    if crds_update:
        setup_crds(ref_dir, verbose)
    
    # Test that $oref is properly defined:
    oref = os.environ.get('oref', failobj='Undefined')
    if oref is 'Undefined' or not os.access(oref, os.R_OK):
        raise OSError('Cannot read $oref directory!\n    {}\n'.format(oref) + \
            '    Please set $oref environmental variable appropriately or run with --crds_update.')
    if verbose:
        print '$oref      = {}\n'.format(oref)
    
    raw_files = determine_input_science(science_dir, allow, verbose)
    log.flush()
    
    # Rename output files according to this:
    output_mapping = {
        'cte_flt.fits' : 'flc.fits' ,
        'cte_crj.fits' : 'crc.fits' ,
        'cte_sx2.fits' : 's2c.fits' ,
        'cte_x2d.fits' : 'x2c.fits' ,
        'cte_sx1.fits' : 's1c.fits' ,
        'cte_x1d.fits' : 'x1c.fits' ,
        'blt_tra.txt'  : 'trb.txt'  ,
        'cte_tra.txt'  : 'trc.txt'  ,
        'blt.fits'     : '<remove>' ,
        'cte.fits'     : '<pass>'   }
    
    # Check for results from previous runs:
    rootnames = [os.path.basename(f).split('_',1)[0] for f in raw_files]
    check_for_old_output_files(rootnames, science_dir, output_mapping, clean, verbose)
    # (Note that the 'clean' option doesn't remove '*.txt' files.)
    
    # Run crds.BestrefsScript on science files:
    if crds_update:
        # Run correction on matching _wav files too, if they exist:
        raw_path = os.path.dirname(raw_files[0])
        wav_files = []
        for file in raw_files:
            with fits.open(file) as f:
                wav_files.append(f[0].header.get('WAVECAL', default='N/A').strip())
        wav_files = [file for file in wav_files if 'N/A' not in file]
        wav_files = [os.path.join(raw_path, file) for file in wav_files]
        wav_files = [file for file in wav_files if os.path.exists(file)]
        
        crds_files = deepcopy(raw_files)
        crds_files.extend(wav_files)
        
        if verbose:
            print 'Running crds.BestrefsScript on science and wav files...'
        errors = BestrefsScript('BestrefsScript --update-bestrefs -s 1 -f ' + ' '.join(crds_files))()
        if int(errors) > 0:
            raise Exception('CRDS BestrefsScript:  Call returned errors!')
    
    # Check science files for uncorrected super-darks:
    populate_darkfiles(raw_files, dark_dir, ref_dir, pctetab, num_processes, all_weeks_flag, 
        clean_all, crds_update, ignore_missing, verbose)
    log.flush()
    
    # Bias-correct the science files:
    bias_corrected = bias_correct_science_files(raw_files, verbose)
    log.flush()
    
    # Perform the CTI correction in parallel on the science data:
    cti_corrected = perform_cti_correction(bias_corrected, pctetab, num_processes, False, verbose)
    log.flush()
    
    # Finish running CalSTIS on the CTI-corrected science data:
    flts = run_calstis_on_science(cti_corrected, verbose)
    log.flush()
    
    # Delete intermediate products and rename final products:
    map_outputs(rootnames, science_dir, output_mapping, verbose)
    log.flush()
    
    end_time = datetime.datetime.now()
    print '\nCompletion time:                {}'.format(end_time.isoformat(' '))
    print 'Run time:                       {}'.format(end_time - start_time)
    print 'stis_cti.py complete!\n'
    log.close()


def setup_crds(ref_dir, verbose=False):
    '''
    Setup $CRDS_PATH and $oref environmental variables to house CRDS reference files.
    
    Sets the $CRDS_PATH environmental variable according to:
        (1) If undefined, Central Store location (if possible)
        (2) If not on Central Store (local), read/writable directory defined by system's $CRDS_PATH
        (3) If local and system's $CRDS_PATH directory not read/writable, nested within 'ref'
    
    Unless $CRDS_PATH is on Central Store, $oref is redefined to be nested within new $CRDS_PATH.
    
    $oref value must end in the path separator (e.g. '/'), so this is appended if necessary.
    '''
    
    if os.environ.get('CRDS_SERVER_URL') is None:
        os.environ['CRDS_SERVER_URL'] = crds_server_url
    
    if verbose:
        print 'Setting up CRDS environmental variables...'
        print '$CRDS_SERVER_URL = {}'.format(os.environ['CRDS_SERVER_URL'])
    
    ref_dir = resolve_iraf_file(ref_dir)
    
    if os.environ.get('CRDS_PATH') is None and \
       os.path.exists('/grp/crds/cache') and os.path.exists('/grp/crds/cache/references/hst'):
            # On-site CRDS cache on Central Store:
            # (This location is special because we can assume it does not need updating.)
            os.environ['CRDS_PATH'] = '/grp/crds/cache'
            os.environ['oref'] = os.path.join(os.environ['CRDS_PATH'], 'references', 'hst') + \
                                 os.path.sep
    elif os.path.exists('/grp/crds/cache') and \
         os.path.samefile(os.environ.get('CRDS_PATH'), '/grp/crds/cache'):
        # CRDS_PATH is already defined as Central Store:
        os.environ['oref'] = os.path.join(os.environ['CRDS_PATH'], 'references', 'hst') + \
                             os.path.sep
    else:
        if os.environ.get('CRDS_PATH') is None:
            # $CRDS_PATH is undefined and Central Store is not available:
            os.environ['CRDS_PATH'] = os.path.abspath(ref_dir)
            os.environ['oref'] = os.path.join(os.environ['CRDS_PATH'], 'references', 'hst', 'stis') + \
                                 os.path.sep
        elif not os.access(os.environ.get('CRDS_PATH'), os.R_OK | os.W_OK):
            # $CRDS_PATH was already defined by the user, but is NOT read/writable:
            print 'WARNING:  Local $CRDS_PATH is not read/writable.'
            print "         Using 'ref' directory for local CRDS cache."
            os.environ['CRDS_PATH'] = os.path.abspath(ref_dir)
            os.environ['oref'] = os.path.join(os.environ['CRDS_PATH'], 'references', 'hst', 'stis') + \
                                 os.path.sep
        elif os.access(os.environ.get('CRDS_PATH'), os.R_OK | os.W_OK):
            # $CRDS_PATH was already defined by the user and is read/writable:
            
            # Check that $oref is defined within $CRDS_PATH properly for either a flat or
            # instrument-specific cache:
            if os.path.abspath(os.environ.get('oref')) not in \
                [os.path.abspath(os.path.join(os.environ.get('CRDS_PATH'), \
                                              'references', 'hst')), 
                 os.path.abspath(os.path.join(os.environ.get('CRDS_PATH'), \
                                              'references', 'hst', 'stis'))]:
                raise OSError('$oref is not properly defined within $CRDS_PATH.\n' + 
                              '    Consider fixing $oref or running without --crds_update option.')
        else:
            raise Exception('Unexpected condition.')
        
        if not os.access(os.path.abspath(os.environ.get('oref')), os.F_OK):
            try:
                os.makedirs(os.path.abspath(os.environ.get('oref')))
                if verbose:
                    print 'Created $oref directory.'
            except OSError as err:
                print 'Could not create directory structure:'
                print '    {}'.format(os.path.abspath(os.environ.get('oref')))
                raise err
        
        # Check that local $CRDS_PATH and $oref are read/writable:
        if not os.access(os.environ.get('CRDS_PATH'), os.R_OK | os.W_OK) or \
           not os.access(os.environ.get('oref'), os.R_OK | os.W_OK):
            raise OSError(('Local $CRDS_PATH and/or $oref are not read/writable!\n' + 
                           '    $CRDS_PATH = {}\n' + 
                           '    $oref      = {}\n' +
                           '    Either fix these or run without --crds_update option.').format( \
                           os.environ.get('CRDS_PATH'), os.environ.get('oref')))
    
    # Check that $oref ends in the path separator:
    if os.environ['oref'][-1] != os.path.sep:
        os.environ['oref'] += os.path.sep
    
    if not os.access(os.environ['oref'], os.R_OK):
        raise OSError('Cannot read $oref!\n    {}'.format(os.environ['oref']))
    
    if verbose:
        print '$CRDS_PATH = {}'.format(os.environ.get('CRDS_PATH'))
    
    return True


def determine_input_science(science_dir, allow=False, verbose=False):
    # Find science rootnames:
    raw_files = glob.glob(os.path.join(science_dir, '*_raw.fits*'))
    if verbose:
        print 'Input _raw.fits files:'
        print '   ' + '\n   '.join(raw_files) + '\n'
    
    if allow:
        # Allow any HST/STIS non-acqs:
        if verbose:
            print 'Leniently filtering input _raw.fits files...'
        filtered_raw_files = [file for file in raw_files if \
            viable_ccd_file( file, \
                earliest_date_allowed = datetime.datetime(1990,1,1,0,0,0), \
                amplifiers_allowed = ['A','B','C','D'], \
                gains_allowed = [1,2,4,8], \
                offsts_allowed = range(9))]
    else:
        # Normal filtering on files:
        filtered_raw_files = [file for file in raw_files if viable_ccd_file(file)]
    
    not_correcting = list(set(raw_files) - set(filtered_raw_files))
    if len(not_correcting) > 0:
        not_correcting.sort()
        print 'WARNING:  Not running the correction on these files:'
        print '   ' + '\n   '.join(not_correcting) + '\n'
    
    # Make sure some files exist:
    if len(filtered_raw_files) == 0:
        raise FileError('No viable STIS CCD files were present in the directory!\n' + \
                        '   ' + science_dir + '\n'                                  + \
                        'Please make sure science data are:\n'                      + \
                        '   -- post-May 01 2009\n'                                  + \
                        '   -- CCDAMP = D\n'                                        + \
                        '   -- CCDGAIN = {1, 4}\n'                                  + \
                        '   -- Not OBSMODE = ACQ*\n'                                + \
                        '   -- Full-frame and not binned')
    
    if verbose:
        print 'Input _raw.fits files being corrected:'
        print '   ' + '\n   '.join(filtered_raw_files) + '\n'
    
    # Check to see if IMPHTTAB is defined for files that will run PHOTCORR:
    missing_imphttab = []
    for file in filtered_raw_files:
        with fits.open(file) as f:
            if (f[0].header.get('PHOTCORR', default='').strip() == 'PERFORM' and \
                f[0].header.get('IMPHTTAB', default=None) is None):
                missing_imphttab.append(file)
    if len(missing_imphttab) > 0:
        raise IOError('These files will fail CalSTIS reduction:\n' + \
                      '   ' + '\n   '.join(missing_imphttab) + \
                      '\nPlease set the IMPHTTAB keyword in files that will perform PHOTCORR.')
    if verbose:
        print 'All files running PHOTCORR have the IMPHTTAB set.\n'
    
    return filtered_raw_files


def viable_ccd_file(file, 
                    earliest_date_allowed=None, 
                    amplifiers_allowed=None, 
                    gains_allowed=None, 
                    offsts_allowed=None):
    '''
    :param earliest_date_allowed: Beginning datetime allowed.
    :param amplifiers_allowed: CCDAMP values allowed.
    :param gains_allowed:  CCDGAIN values allowed.
    :param offsts_allowed: CCDOFFST values allowed
    
    :type earliest_date_allowed: datetime.datetime object (default=2009-May-01 00:00:00 UTC)
    :type amplifiers_allowed: list of strings (default=['D'])
    :type gains_allowed: list of ints (default=[1,4])
    :type offsts_allowed: list of ints (default=[3])
    
    :returns:  bool -- Run stis_cti.stis_cti() on the file?
    '''
    
    # Set defaults:
    if earliest_date_allowed is None:
        earliest_date_allowed = datetime.datetime(2009, 5, 1, 0, 0, 0)
    if type(earliest_date_allowed) is not datetime.datetime:
        raise TypeError('earliest_date_allowed must be a datetime.datetime, not a {}.'.format(
            type(earliest_date_allowed)))
    
    if amplifiers_allowed is None:
        amplifiers_allowed = ['D']
    
    if gains_allowed is None:
        gains_allowed = [1, 4]
    
    if offsts_allowed is None:
        offsts_allowed = [3]
    
    # Read ext=0 header of observation:
    with fits.open(file) as f:
        hdr0 = f[0].header
    
    # Parse date/time of observation:
    date = hdr0['TDATEOBS'].strip()
    time = hdr0['TTIMEOBS'].strip()
    dt = datetime.datetime.strptime(date + ' ' + time, '%Y-%m-%d %H:%M:%S')
    
    return \
        hdr0['TELESCOP'].strip() == 'HST'              and \
        hdr0['INSTRUME'].strip() == 'STIS'             and \
        hdr0['DETECTOR'].strip() == 'CCD'              and \
        dt >= earliest_date_allowed                    and \
        'ACQ' not in hdr0['OBSMODE'].strip()           and \
        hdr0['CCDAMP'].strip() in amplifiers_allowed   and \
        hdr0['CCDGAIN'] in gains_allowed               and \
        hdr0['CCDOFFST'] in offsts_allowed             and \
        not hdr0['SUBARRAY']                           and \
        hdr0['BINAXIS1'] == 1 and hdr0['BINAXIS2'] == 1


def bias_correct_science_files(raw_files, verbose=False):
    if verbose:
        print 'Bias-correcting science files...\n'
    
    outnames = [f.replace('_raw.fits', '_blt.fits', 1) for f in raw_files]
    trailers = [os.path.abspath(os.path.expandvars(f.replace('_raw.fits', '_blt_tra.txt', 1))) for f in raw_files]
    
    # Check for previous _blt.fits files first:
    for outname in outnames:
        if os.path.exists(outname):
            raise IOError('File {} already exists!'.format(outname))
    
    for raw_file, outname, trailer in zip(raw_files, outnames, trailers):
        if verbose:
            print 'Running basic2d on {} --> {}.'.format(raw_file, outname)
        if os.path.exists(trailer):
            os.remove(trailer)
        cwd = os.getcwd()
        try:
            # Need to change to science directory to find associated epc files.
            os.chdir(os.path.dirname(raw_file))
            status = basic2d.basic2d(os.path.basename(raw_file), output=os.path.basename(outname), 
                dqicorr='perform', atodcorr='omit', blevcorr='perform', biascorr='perform', 
                doppcorr='omit', lorscorr='omit', glincorr='omit', lflgcorr='omit', 
                darkcorr='omit', flatcorr='omit', shadcorr='omit', photcorr='omit', 
                statflag=True, verbose=(verbose >= 2), timestamps=False, trailer=trailer)
        finally:
            os.chdir(cwd)
        
        if verbose or status != 0:
            with open(os.path.expandvars(trailer)) as tra:
                for line in tra.readlines():
                    print '     ' + line.strip()
            print
        
        if status != 0:
            raise RuntimeError('basic2d returned non-zero status on {}:  {}'.format(raw_file, status))
    
    return outnames


def run_calstis_on_science(files, verbose):
    if verbose:
        print 'Running CalSTIS on science files...\n'
    
    # Note that outname is determined by CalSTIS, and not this function.
    outnames = [f.replace('_cte.fits', '_cte_flt.fits', 1) for f in files]  # Replicating CalSTIS' behavior
    trailers = [os.path.abspath(os.path.expandvars(f.replace('_cte.fits', '_cte_tra.txt', 1))) for f in files]
    
    # Check for previous output files:
    for outname in outnames:
        if os.path.exists(outname):
            raise IOError('File {} already exists!'.format(outname))
    
    for file, outname, trailer in zip(files, outnames, trailers):
        if verbose:
            print 'Running calstis on {} --> {}.'.format(file, outname)
        if os.path.exists(trailer):
            os.remove(trailer)
        
        cwd = os.getcwd()
        try:
            # Need to change to science directory to find associated wavecals.
            os.chdir(os.path.dirname(file))
            status = calstis.calstis(os.path.basename(file), verbose=(verbose >= 2), 
                timestamps=False, trailer=trailer)
        finally:
            os.chdir(cwd)
        
        if verbose or status != 0:
            with open(os.path.expandvars(trailer)) as tra:
                for line in tra.readlines():
                    print '     ' + line.strip()
            print
        
        if status != 0:
            raise RuntimeError('CalSTIS returned non-zero status on {}:  {}'.format(file, status))
    
    return outnames


def resolve_iraf_file(file):
    '''Resolves pathvar$filename into usable format.'''
    dir = ''
    rootname = file
    
    if '$' in file and file[0] == '$':
        dir, rootname = file[1:].split(os.path.sep, 1)
    elif '$' in file:
        dir, rootname = file.split('$', 1)
    
    if dir != '':
        dir_resolved = os.getenv(dir)
        if dir_resolved is None:
            raise IOError('Can\'t resolve environmental variable \'{}\'.'.format(dir))
        else:
            dir = dir_resolved
    
    new_file = os.path.normpath(os.path.join(dir, rootname))
    
    return new_file


def check_pctetab_version(pctetab, verbose=False, version_min='0.1', version_max='1.999'):
    '''
    Make sure the version keyword in the PCTETAB is within the acceptable boundaries
    for this version of the stis_cti.py code.
    
    This comparison works for version = "<int>.<int>" (e.g. "1.10" > "1.1").
    '''
    if type(version_min) != str or type(version_max) != str:
        raise TypeError('Versions must be strings.')
    
    # Handle environmental variables in the PCTETAB:
    pctetab = os.path.normpath(resolve_iraf_file(pctetab))
    
    # Read header keyword VERSION and strip anything after '_':
    with fits.open(pctetab) as p:
        pctetab_version_raw = p[0].header.get('PCTE_VER', default='').strip()
        pctetab_version = pctetab_version_raw.split('_')[0]
    
    if pctetab_version == '':
        raise VersionError('PCTE_VER keyword not found in PCTETAB {}.'.format(pctetab))
    
    pctetab_version_major = int(pctetab_version.split('.')[0])
    pctetab_version_minor = int(pctetab_version.split('.')[1])
    
    if (pctetab_version_major   < int(version_min.split('.')[0])  or
        (pctetab_version_major == int(version_min.split('.')[0]) and
         pctetab_version_minor  < int(version_min.split('.')[1]))):
        raise VersionError(('Version mismatch between PCTETAB {} and stis_cti.py code.\n' + 
            'Please download a more recent version of this reference file!\n').format(pctetab))
       
    if (pctetab_version_major   > int(version_max.split('.')[0])  or
        (pctetab_version_major == int(version_max.split('.')[0]) and
         pctetab_version_minor  > int(version_max.split('.')[1]))):
        raise VersionError(('Version mismatch between PCTETAB {} and stis_cti.py code.\n' +
            'Please upgrade this code!\n').format(pctetab))\
    
    if verbose:
        print 'Version of PCTETAB:  {}\n'.format(pctetab_version_raw)
    
    return True


def superdark_hash(sim_nit=None, shft_nit=None, rn_clip=None, nsemodel=None, subthrsh=None, pcte_ver=None,
                            pctetab=None,
                            superdark=None, files=None):
    from numpy import where
    
    if pctetab is not None:
        with fits.open(resolve_iraf_file(pctetab)) as f:
            sim_nit  = f[0].header['SIM_NIT' ]
            shft_nit = f[0].header['SHFT_NIT']
            rn_clip  = f[0].header['RN_CLIP' ]
            nsemodel = f[0].header['NSEMODEL']
            subthrsh = f[0].header['SUBTHRSH']
            pcte_ver = f[0].header['PCTE_VER'].strip()
        if files is None:
            raise IOError('Must specify file list with pctetab.')
    elif superdark is not None:
        with fits.open(os.path.expandvars(superdark)) as f:
            sim_nit  = f[0].header.get('PCTESMIT', default='undefined')
            shft_nit = f[0].header.get('PCTESHFT', default='undefined')
            rn_clip  = f[0].header.get('PCTERNCL', default='undefined')
            nsemodel = f[0].header.get('PCTENSMD', default='undefined')
            subthrsh = f[0].header.get('PCTETRSH', default='undefined')
            pcte_ver = f[0].header.get('PCTE_VER', default='0.0').strip()
            if files is None:
                # Parse superdark header for files used:
                hist = f[0].header['HISTORY']
                beginning = where(['The following input files were used:' in line for line in hist])[0][0] + 1
                end = where([line == '' for line in hist[beginning:]])[0][0] + beginning
                files = hist[beginning:end]
    else:
        if sim_nit  == None or \
           shft_nit == None or \
           rn_clip  == None or \
           nsemodel == None or \
           subthrsh == None or \
           pcte_ver == None:
            raise IOError('Please specify either pctetab or the proper set of parameters!')
        if files is None:
            raise IOError('Must specify files with pctetab.')
    
    # Ignore PCTETAB version text after underscore:
    pcte_ver = pcte_ver.split('_',1)[0]
    
    # Turn list of filenames into a sorted list of exposures:
    exposures = [os.path.basename(file).rsplit('_',1)[0].upper() for file in files]
    exposures.sort()
    exposures = ','.join(exposures)
    
    hash_str = '{};{};{};{};{};{};{}'.format( \
        exposures, \
        sim_nit, shft_nit, rn_clip, nsemodel, subthrsh, pcte_ver)
    
    return hash(hash_str)


# These functions are needed here for calling from a multiprocessing pool.
def func(in_f, out_f):
    StisPixCteCorr.CteCorr(in_f, outFits=out_f)

def func_star(a_b):
    """Convert `f([1,2])` to `f(1,2)` call."""
    return func(*a_b)


def perform_cti_correction(files, pctetab, num_cpu=1, clean_all=False, verbose=False):
    '''Run StisPixCteCorr when needed.'''
    
    perform_files = []
    outnames = []
    perform_outnames = []
    for file in files:
        outname = file.replace('_flt.fits', '_cte.fits', 1).replace('_blt.fits', '_cte.fits', 1)
        outnames.append(outname)
        
        if clean_all and os.path.exists(outname):
            if verbose:
                print 'Deleting file:  {}'.format(outname)
            os.remove(outname)
        
        if os.path.exists(outname) and (superdark_hash(pctetab=pctetab, files=[]) == superdark_hash(superdark=outname, files=[])):
            if verbose:
                print 'Skipping regeneration of CTI-corrected file:  {}'.format(outname)
        else:
            with fits.open(file, 'update') as f:
                #Update PCTETAB header keyword:
                old_pctetab = f[0].header.get('PCTETAB', default=None)
                f[0].header.set('PCTETAB', pctetab, 'Pixel-based CTI param table', after='DARKFILE')
                
                # Turn off the empirical CTI correction flag if it is set to PERFORM:
                old_ctecorr = f[0].header.get('CTECORR', default='unknown').strip()
                if old_ctecorr == 'COMPLETE':
                    raise ValueError('Empirical CTI correction correction flag, CTECORR, is set in file {}.'.format(file))
                elif old_ctecorr == 'PERFORM':
                    f[0].header['CTECORR'] = 'OMIT'
                
                f.flush()
                if verbose:
                    print 'Updated hdr0 PCTETAB  of {}:  {} --> {}'.format(file, old_pctetab, pctetab)
                    if old_ctecorr != 'unknown':
                        print 'Updated hdr0 CTECORR  of {}:  {} --> {}'.format(
                            file, old_ctecorr, f[0].header['CTECORR'])
                
                # Run the pixel-based correction on these files:
                perform_files.append(file)
                perform_outnames.append(outname)
    
    # Run the CTI-correction:
    p = multiprocessing.Pool(processes = num_cpu)
    # Iterator of input/output files:
    file_args = ((x, y) for x, y in zip(perform_files, perform_outnames))
    p.map_async(func_star, file_args)
    p.close()
    p.join()
    
    # Single-threaded version:
    #for perform_file, outname in zip(perform_files, outnames):
    #    StisPixCteCorr.CteCorr(perform_file, outFits=outname)
    
    if len(outnames) == 0:
        outnames = None
    
    return outnames


def copy_dark_keywords(superdark, dark_hdr0, pctetab, history=None, basedark=None):
    '''Copy header keywords from a used component dark to the new super-dark.'''
    
    # Copy these keywords from the last component dark to the new superdark:
    keywords = ['PCTECORR', 'PCTETAB', 'PCTEFRAC', 'PCTERNCL', 'PCTERNCL', 'PCTENSMD', 
                'PCTETRSH', 'PCTESMIT', 'PCTESHFT', 'CTE_NAME', 'CTE_VER', 'PCTE_VER']
    # Note:  PCTEFRAC is time-dependent; PCTECORR = COMPLETE
    keywords.reverse()
    
    with fits.open(os.path.expandvars(superdark), 'update') as s:
        for keyword in keywords:
            value = dark_hdr0.get(keyword, default='unknown')
            try:
                comment = dark_hdr0.comments[keyword]
            except KeyError:
                comment = None
            
            if keyword in s[0].header:
                s[0].header[keyword] = value
            else:
                s[0].header.set(keyword, value, comment, after='DRK_VS_T')
            s.flush()
        
        if basedark is not None:
            if 'BASEDARK' in s[0].header:
                s[0].header['BASEDARK'] = basedark
            else:
                s[0].header.set('BASEDARK', basedark, 'Used to make weekdark', after=keywords[0])
        
        # Add HISTORY to superdark hdr0 here:
        if history is not None:
            s[0].header['HISTORY'] = ' '
            for h in history:
                s[0].header['HISTORY'] = h


def generate_basedark(files, outname, pctetab, num_cpu, clean_all=False, verbose=False):
    '''Generate a basedark for an annealing period, if it doesn't already exist.'''
    
    if os.path.exists(os.path.expandvars(outname)):
        if clean_all:
            if verbose:
                print 'Deleting basedark:  {}'.format(os.path.expandvars(outname))
            os.remove(os.path.expandvars(outname))
        elif superdark_hash(pctetab=pctetab, files=files) == superdark_hash(superdark=outname):
            # Don't make a basedark if it already exists:
            if verbose:
                print 'Skipping regeneration of basedark:  {}'.format(outname)
            return
    
    if verbose >= 2:
        print 'Working on basedark {}:'.format(outname)
        print '   ' + '\n   '.join(files) + '\n'
    
    # Correct component darks, if necessary:
    corrected_files = perform_cti_correction(files, pctetab, num_cpu, clean_all, verbose)
    
    # Make a basedark from the corrected darks:
    refstis.basedark.make_basedark(corrected_files, refdark_name=os.path.normpath(os.path.expandvars(outname)))
    
    if verbose:
        calstis_log = outname.replace('.fits','_joined_bd_calstis_log.txt', 1)
        print 'Calstis log file for CRJ processing [{}]:'.format(calstis_log)
        with open(os.path.expandvars(calstis_log), 'r') as cs_log:
            for line in cs_log.readlines():
                print '     ' + line.strip()
        print
    
    # Copy the last file's ext=0 header into a variable to use in populating the basedark header:
    with fits.open(corrected_files[-1]) as file:
        dark_hdr0 = file[0].header
    
    # Update keywords in new basedark from component dark:
    history = [
        'Basedark created from CTI-corrected component darks by script',
        'stis_cti.py on {}.'.format(datetime.datetime.now().isoformat(' ')) ]
    copy_dark_keywords(os.path.expandvars(outname), dark_hdr0, pctetab, history=history)
        
    if verbose:
        print 'Basedark complete:  {}\n'.format(outname)


def generate_weekdark(files, outname, pctetab, basedark, num_cpu, clean_all=False, verbose=False):
    '''Generate a weekdark for part of an annealing period, if it doesn't already exist.'''
    
    if os.path.exists(os.path.expandvars(outname)):
        if clean_all:
            if verbose:
                print 'Deleting weekdark:  {}'.format(os.path.expandvars(outname))
            os.remove(os.path.expandvars(outname))
        elif superdark_hash(pctetab=pctetab, files=files) == superdark_hash(superdark=outname):
            # Don't make a weekdark if it already exists:
            if verbose:
                print 'Skipping regeneration of weekdark:  {}'.format(outname)
            return
    
    if verbose >= 2:
        print 'Working on weekdark {}:'.format(outname)
        print '   ' + '\n   '.join(files) + '\n'
    
    # Correct component darks, if necessary:
    corrected_files = perform_cti_correction(files, pctetab, num_cpu, False, verbose)
      # Don't delete component darks here, as this has already been done in generate_basedark().
    
    # Make a weekdark from the corrected darks:
    refstis.weekdark.make_weekdark(corrected_files, os.path.normpath(os.path.expandvars(outname)), 
        os.path.abspath(os.path.expandvars(basedark)))
    
    if verbose:
        calstis_log = outname.replace('.fits','_joined_bd_calstis_log.txt', 1)
        print 'Calstis log file for CRJ processing [{}]:'.format(calstis_log)
        with open(os.path.expandvars(calstis_log), 'r') as cs_log:
            for line in cs_log.readlines():
                print '     ' + line.strip()
        print
    
    # Copy the last file's ext=0 header into a variable to use in populating the basedark header:
    with fits.open(corrected_files[-1]) as file:
        dark_hdr0 = file[0].header
    
    # Update keywords in new basedark from component dark:
    history = [
        'Weekdark created from CTI-corrected component darks by script',
        'stis_cti.py on {}'.format(datetime.datetime.now().isoformat(' ')),
        'using basedark file {}.'.format(basedark)]
    copy_dark_keywords(os.path.expandvars(outname), dark_hdr0, pctetab, history=history, basedark=basedark)
    
    if verbose:
        print 'Weekdark complete:  {}\n'.format(outname)


def populate_darkfiles(raw_files, dark_dir, ref_dir, pctetab, num_processes, all_weeks_flag=False, 
    clean_all=False, crds_update=False, ignore_missing=False, verbose=False):
    '''Check science files for uncorrected super-darks; and, if necessary, generate them and
       populate the science file headers.'''
    
    superdark_remakes = []
    for file in raw_files:
        with fits.open(file) as f:
            hdr0 = f[0].header
        superdark = hdr0['DARKFILE']
        superdark_resolved = resolve_iraf_file(superdark)
        try:
            with fits.open(superdark_resolved) as sd:
                hdr0 = sd[0].header
                if hdr0.get('PCTECORR', default='unknown').strip().upper() == 'COMPLETE':
                    if not clean_all:
                        if verbose:
                            print 'Superdark {} is already CTI-corrected.'.format(superdark_resolved)
                        pass
                    else:
                        if verbose:
                            print 'Remaking CTI-corrected superdark {}'.format(superdark_resolved)
                        if os.path.abspath(os.path.dirname(superdark_resolved)) == os.path.abspath(ref_dir):
                            os.remove(superdark_resolved)
                            if verbose:
                                print 'Deleted old superdark:  {}'.format(superdark_resolved)
                        superdark_remakes.extend(file)
                else:
                    if verbose:
                        print 'Superdark {} is not CTI-corrected.'.format(superdark_resolved)
                    superdark_remakes.extend(file)
        except IOError:
            if verbose:
                print 'Superdark {} not found!  Will make a CTI-corrected version.'.format(superdark_resolved)
            superdark_remakes.extend(file)
    
    if len(superdark_remakes) == 0:
        if verbose:
            print 'Not remaking any superdarks.\n'
        return  # Nothing to remake!
    
    if verbose:
        print 'Making superdarks for:'
        print '   ' + '\n   '.join(raw_files) + '\n'
    
    # Determine component darks used to make superdarks:
    anneal_data = archive_dark_query.get_anneal_boundaries()
    anneals = archive_dark_query.archive_dark_query( \
                      raw_files, anneal_data=anneal_data, print_url=False)
    
    # Get list of EXPNAMEs from files in the dark_dir.
    found_dark_files = glob.glob(os.path.join(dark_dir, '*_flt.fits*'))
    if verbose >= 2:
        max_file_length = str(max(map(lambda x: len(x), found_dark_files) + [1]))
    
    found = {}
    for file in found_dark_files:
        with fits.open(file) as f:
            hdr0 = f[0].header
        found[hdr0['ROOTNAME'].strip().upper()] = (file, hdr0)
    
    # Loop over expected darks within anneals and populate keyword data, noting missing files:
    get_keywords = ['CCDAMP', 'CCDGAIN', 'CCDOFFST', 'BIASFILE', 'DARKFILE', 'TELESCOP', 'INSTRUME', 'DETECTOR']
    missing_darks = set()
    weekdark_types = defaultdict(list)
    for anneal in anneals:
        # Populate file locations:
        for dark in anneal['darks']:
            if found.has_key(dark['exposure']):
                dark['file'] = found[dark['exposure']][0]
            else:
                missing_darks.add(dark['exposure'])
        
        # Update file headers:
        if crds_update:
            dark_files_to_run = [d['file'] for d in anneal['darks'] if d.has_key('file')]
            if len(dark_files_to_run) > 0:
                if verbose:
                    print 'Running crds.BestrefsScript on dark files from anneal {}...'.format(anneal)
                
                errors = BestrefsScript('BestrefsScript --update-bestrefs -s 1 -f ' + ' '.join(dark_files_to_run))()
                if int(errors) > 0:
                    raise Exception('CRDS BestrefsScript (darks):  Call returned errors!')
                
                # Update hdr0 data after BestrefsScript is run:
                for dark in anneal['darks']:
                    if dark.has_key('file'):
                        file = dark['file']
                        with fits.open(file) as f:
                            hdr0 = f[0].header
                        found[hdr0['ROOTNAME'].strip().upper()] = (file, hdr0)
        
        # For each amp, count weeks based on old darkfile; reset for each annealing period:
        # Get keywords about each dark:
        all_weeks = {'A':{}, 'B':{}, 'C':{}, 'D':{}}
        for dark in anneal['darks']:
            if found.has_key(dark['exposure']):
                for get_keyword in get_keywords:
                    dark[get_keyword] = found[dark['exposure']][1][get_keyword]
                    try:
                        # Strip leading/trailing spaces off of strings:
                        dark[get_keyword] = dark[get_keyword].strip()
                    except AttributeError:
                        pass
                
                # Determine week number (for weekdark separation) for each amp:
                if len(all_weeks[dark['CCDAMP']]) == 0:
                    all_weeks[dark['CCDAMP']][dark['DARKFILE']] = 1
                elif dark['DARKFILE'] not in all_weeks[dark['CCDAMP']].keys():
                    all_weeks[dark['CCDAMP']][dark['DARKFILE']] = len(all_weeks[dark['CCDAMP']]) + 1
                dark['week_num'] = all_weeks[dark['CCDAMP']][dark['DARKFILE']]
                # Construct a unique tag for each weekdark and assign it to the component darks:
                dark['weekdark_tag'] = '{}{:03.0f}_{:1.0f}'.format(dark['CCDAMP'].lower(), anneal['index'], dark['week_num'])
                weekdark_types[dark['weekdark_tag']].append(dark)
                
                if verbose >= 2:
                    print ('{:s} {:5s} {:' + max_file_length + 's} {} {} {} {} {:5.0f} {} {} {}').format( \
                        dark['exposure'], dark['DETECTOR'], dark['file'], dark['CCDAMP'], \
                        dark['CCDGAIN'], dark['CCDOFFST'], dark['datetime'].isoformat(' '), \
                        dark['exptime'], dark['week_num'], dark['weekdark_tag'], dark['DARKFILE'])
            else:
                missing_darks.add(dark['exposure'])
    if verbose >= 2:
        print
    
    if len(missing_darks) != 0:
        missing_darks = list(missing_darks)
        missing_darks.sort()
        print 'ERROR:  These FLT component darks are missing from {}:'.format(dark_dir)
        print ', '.join(missing_darks) + '\n'
        if not ignore_missing:
            print 'Please download the missing darks (calibrated FLTs) via this link:'
            print '(or specify the proper dark_dir [{}])\n'.format(dark_dir)
            print 'If missing files are expected (e.g. amp=A darks), then run with'
            print '--ignore_missing flag.\n'
            print archive_dark_query.darks_url(missing_darks) + '\n'
            sys.exit(1)
        else:
            print '"--ignore_missing" flag set:  Ignoring these missing FLT files...'
    elif verbose:
        print 'All required component dark FLT files for annealing periods have been located on disk.\n'
    
    # Determine week_num time boundaries (approximate for now):
    weekdarks = {}
    for weekdark_tag in weekdark_types:
        weekdarks[weekdark_tag] = {
            'amp'          : weekdark_tag[0].upper(), 
            'start'        : min([x['datetime'] for x in weekdark_types[weekdark_tag]]), 
            'end'          : max([x['datetime'] for x in weekdark_types[weekdark_tag]]), 
            'darks'        : list(weekdark_types[weekdark_tag]), 
            'anneal_num'   : int(weekdark_tag[1:4]), 
            'week_num'     : int(weekdark_tag[5]), 
            'weekdark_tag' : weekdark_tag }
    
    # Loop over each {amp, anneal_num} combination and adjust week boundaries:
    amps_anneals = list(set([(weekdarks[x]['amp'], weekdarks[x]['anneal_num']) for x in weekdarks]))
    for amp_anneal in amps_anneals:
        # Determine which data matches the (amp, anneal) specified in loop iteration:
        amp_anneal_data = [weekdarks[x] for x in weekdarks 
            if weekdarks[x]['amp'] == amp_anneal[0] and weekdarks[x]['anneal_num'] == amp_anneal[1]]
        weeks = list(amp_anneal_data)
        weeks.sort(key=lambda x: x['week_num'])
        # Extend start/end boundaries to anneal boundaries:
        anneal = filter(lambda x: weeks[0]['anneal_num'] == x['index'], anneals)[0]
        weekdarks[weeks[ 0]['weekdark_tag']]['start'] = anneal['start']  # start of annealing period
        weekdarks[weeks[-1]['weekdark_tag']]['end']   = anneal['end']    # end of annealing period
        # Make inter-week boundaries contiguous:
        if len(weeks) >= 2:
            for i, week in enumerate(weeks[0:-1]):
                delta = datetime.timedelta(seconds = round((weeks[i+1]['start'] - weeks[i]['end']).total_seconds() / 2.0))
                midpt = weeks[i]['end'] + delta
                weekdarks[weeks[i  ]['weekdark_tag']]['end']   = midpt
                weekdarks[weeks[i+1]['weekdark_tag']]['start'] = midpt
    
    if verbose >= 2:
        # This should be after filtering by amp...
        print 'Could make weekdarks for:'
        for weekdark_tag in weekdarks:
            print '   {}:  {} - {}  [{}]'.format(
                weekdark_tag, 
                weekdarks[weekdark_tag]['start'], 
                weekdarks[weekdark_tag]['end'], 
                len(weekdarks[weekdark_tag]['darks']) )
        print
    
    weekdark = defaultdict(list)
    for file in raw_files:
        with fits.open(file, 'update') as f:
            hdr0 = f[0].header
            
            dt = datetime.datetime.strptime( \
                hdr0['TDATEOBS'].strip() + ' ' + hdr0['TTIMEOBS'].strip(), '%Y-%m-%d %H:%M:%S')
            matched_weekdark_tags = [wd['weekdark_tag'] for wd in weekdarks.values() if \
                wd['start'] <= dt and wd['end'] > dt and \
                wd['amp'] == hdr0['CCDAMP']]
            if len(matched_weekdark_tags) == 0:
                raise IOError(('No matching component darks found to make needed weekdark for amp {}\n' + \
                    'within annealing period boundaries corresponding to science set {}.')\
                    .format(hdr0['CCDAMP'], os.path.basename(file)))
            weekdark_tag = matched_weekdark_tags[0]
            weekdark[weekdark_tag].append(file)
            
            # Update ext=0 hdr of science file:
            # DARKFILE:
            darkfile = os.path.join(ref_dir, weekdark_tag + '_drk.fits')
            old_darkfile = f[0].header['DARKFILE']
            f[0].header['DARKFILE'] = darkfile
            
            f.flush()
            if verbose:
                print 'Updated hdr0 DARKFILE of {}:  {} --> {}'.format(file, old_darkfile, darkfile)
    
    if verbose:
        print '\nWeekdarks needed:'
        for weekdark_tag in weekdark.keys():
            print '   {} [{}]:'.format(weekdark_tag, len(weekdarks[weekdark_tag]['darks']))
            print '      ' + '\n      '.join(weekdark[weekdark_tag])
        print
    
    if all_weeks_flag:
        weekdark_tags = weekdarks.keys()
        if verbose:
            print 'Generating superdarks for all amps/weeks available.\n'
    else:
        weekdark_tags = weekdark.keys()  # Note the missing 's'
    
    # Generate specified basedarks and weekdarks, based on weekdark_tags:
    already_deleted = set()
    for weekdark_tag in weekdark_tags:
        # Make basedark:
        amp = weekdarks[weekdark_tag]['amp'].strip().lower()
        basedark = os.path.join(ref_dir, 'basedark_{}{}_drk.fits'.format( \
            weekdarks[weekdark_tag]['amp'].strip().lower(), weekdarks[weekdark_tag]['anneal_num']))
        anneal = [a for a in anneals if a['index'] == weekdarks[weekdark_tag]['anneal_num']][0]
        files = [f['file'] for f in anneal['darks'] if f.get('CCDAMP') == weekdarks[weekdark_tag]['amp']]
        if clean_all and basedark not in already_deleted:
            clean_this = True
            already_deleted.add(basedark)
        else:
            clean_this = False
        generate_basedark(files, basedark, pctetab, num_processes, clean_this, verbose)
        
        # Make weekdark:
        weekdark_name = os.path.abspath(os.path.expandvars(os.path.join(ref_dir, weekdark_tag + '_drk.fits')))
        files = [f['file'] for f in weekdarks[weekdark_tag]['darks']]  # Already selected for amp
        if clean_all and weekdark_name not in already_deleted:
            clean_this = True
            already_deleted.add(weekdark_name)
        else:
            clean_this = False
        generate_weekdark(files, weekdark_name, pctetab, basedark, num_processes, clean_this, verbose)


def check_for_old_output_files(rootnames, science_dir, output_mapping, clean=False, verbose=False):
    # Combine science path with rootnames:
    path_rootnames = [os.path.join(science_dir, r) for r in rootnames]
    
    # Combine rootnames with temporary and final output names:
    old_files = []
    for ext in output_mapping.keys():
        old_files.extend([f + '_' + ext for f in path_rootnames
                          if '.txt' not in ext])
        old_files.extend([f + '_' + output_mapping[ext] for f in path_rootnames
                          if '<' not in output_mapping[ext] and '.txt' not in output_mapping[ext]])
    
    # Include gzipped versions of filenames:
    tmp = [f + '.gz' for f in old_files]
    tmp.extend([f + '.GZ' for f in old_files])
    
    # Check for the existence of these files:
    error_files = []
    old_files.extend(tmp)
    for old_file in old_files:
        if os.path.exists(old_file) or os.path.islink(old_file):
            if clean:
                if verbose:
                    print 'Removing old file:  {}'.format(old_file)
                os.remove(old_file)
            else:
                error_files.append(old_file)
    if len(error_files) > 0:
        error_files.sort()
        raise IOError('Files exist from previous run of stis_cti.py:  \n{}'.format( \
                      '   ' + '\n   '.join(error_files) + \
                      '\nYou might consider running with the \'--clean\' option specified.'))
    if verbose:
        print 'The science_dir is clear of files from previous runs.\n'
    
    return True


def map_outputs(rootnames, science_dir, output_mapping, verbose=False):
    '''
    Rename outputs according to output_mapping dictionary.
    
    Special commands for output products:
        '<pass>'   -- Don't do anything with product; denotes it as an intermediate product.
        '<remove>' -- Delete the intermediate product.
    '''
    
    for rootname in rootnames:
        if verbose:
            print 'Renaming files with rootname {}:'.format(rootname)
        files = glob.glob(os.path.join(science_dir, rootname + '*'))
        for file in files:
            cwd = os.getcwd()
            try:
                os.chdir(os.path.dirname(file))
                file = os.path.basename(file)
                
                # Determine file extension:
                try:
                    ext = file.split('_',1)[1]
                except IndexError:
                    ext = '<undefined>'  # No underscore in name
                
                # Handle gzipped filenames:
                if ext[-3:] in ['.gz', '.GZ']:
                    gzip = ext[-3:]
                    ext = ext[:-3]
                else:
                    gzip = ''
                
                # Map action to file:
                if ext not in output_mapping.keys():
                    pass
                elif output_mapping[ext] == '<pass>':
                    pass
                elif output_mapping[ext] == '<remove>':
                    if verbose:
                        print 'Deleting:  {}'.format(file)
                    os.remove(file)
                else:
                    new_file = rootname + '_' + output_mapping[ext] + gzip
                    if verbose:
                        print 'Renaming:  {}\t-->\t{}'.format(file, new_file)
                    os.rename(file, new_file)
            finally:
                os.chdir(cwd)


class Logger(object):
    '''Copies sys.stdout to a log file
       source: http://stackoverflow.com/a/24583265
       Modified to include STDERR.
    '''
    def __init__(self, filename='cti_{}.log'.format(datetime.datetime.now().isoformat('_')), mode="a", buff=0, disable=False):
        self.disable = disable
        
        if not self.disable:
            self.stdout = sys.stdout
            self.stderr = sys.stderr
            self.file = open(filename, mode, buff)
            sys.stdout = self
            sys.stderr = self
    
    def __del__(self):
        self.close()
    
    def __enter__(self):
        pass
    
    def __exit__(self, *args):
        pass
    
    def write(self, message):
        if not self.disable:
            self.stdout.write(message)  # Both STDOUT and STDERR get directed to STDOUT!
            self.file.write(message)
    
    def flush(self):
        if not self.disable:
            self.stdout.flush()
            self.stderr.flush()
            self.file.flush()
            os.fsync(self.file.fileno())
    
    def close(self):
        if not self.disable:
            if self.stdout != None:
                sys.stdout = self.stdout
                self.stdout = None
            
            if self.stderr != None:
                sys.stderr = self.stderr
                self.stderr = None
            
            if self.file != None:
                self.file.close()
                self.file = None



'''
Usage:
import stis_cti
stis_cti.stis_cti(science_dir, dark_dir, ref_dir, num_processes, pctetab=None, 
                  all_weeks_flag=False, allow=False, clean=False, clean_all=False, verbose=1)
'''

from stis_cti import stis_cti, __version__, __author__
from stis_cti import viable_ccd_file
from custom_superdark_info import custom_superdark_info
import StisPixCteCorr
from archive_dark_query import archive_dark_query



def custom_superdark_info():
    '''
    Prints informative text on manually applying the CTI-correction and creating custom
    super-darks.
    '''
    import stis_cti
    import pkg_resources
    import glob
    import os
    
    # Find the stis_cti package's most recent PCTETAB:
    pctetabs = glob.glob(pkg_resources.resource_filename(stis_cti.__name__, 'data/*_pcte.fits'))
    pctetabs.sort()
    if len(pctetabs) > 0:
        pctetab = pctetabs[-1]
    pctetab_dir = os.path.dirname(pctetab) + os.path.sep
    pctetab_file = os.path.basename(pctetab)
    
    print 'How to create and apply custom super-darks'
    print '------------------------------------------\n'
    print 'See the documentation at:  http://pythonhosted.org/stis_cti/readme.html#custom-super-darks\n'
    
    print 'The current stis_cti package PCTETAB reference file is:\n    {}\n'.format(pctetab)
    
    print 'To set the $pctetab environmental variable in the shell:'
    print '    In TCSH:  setenv pctetab {}'.format(pctetab_dir)
    print '    In BASH:  export pctetab="{}"'.format(pctetab_dir)
    print "    Within Python:"
    print "        import os"
    print "        os.environ['pctetab'] = '{}'\n".format(pctetab_dir)
    
    print 'In Python, set the PCTETAB header keyword in the appropriate files to be corrected:'
    print '    from astropy.io import fits\n'
    print "    fits.setval('darks/filename_flt.fits', 'DARKFILE', value='pctetab${}')".format(pctetab_file)
    print '    -- or --'
    print "    fits.setval('science/filename_raw.fits', 'DARKFILE', value='pctetab${}')\n".format(pctetab_file)
    
    print 'Now you can run stis_cti.StisPixCteCorr.CteCorr() on the file to produce a _cte.fits intermediate file:'
    print '    import stis_cti'
    print "    stis_cti.StisPixCteCorr.CteCorr('darks/filename_flt.fits')\n"
    
    print '(Of course, you probably want to loop over all files of interest.)\n'
    
    print 'Component darks corrected this way can be fed into refstis.basedark.make_basedark() and '
    print 'refstis.weekdark.make_weekdark() in order to make custom super-darks.\n'
    print 'First, manually separate the darks within the anneal month of interest and the week'
    print 'surrounding the science files.  Then, run:'
    
    print "    import refstis"
    print "    import glob\n"
    print "    month_files = glob.glob('annealing_month/*_cte.fits')"
    print "        # Assuming only the annealing month's darks are selected"
    print "    refstis.basedark.make_basedark(month_files, refdark_name='basedark_drk.fits')\n"
    print "    week_files = glob.glob('my_week/*_cte.fits')  # Where we have moved the week's files to my_week/"
    print "    refstis.weekdark.make_weekdark(week_files, refdark_name='weekdark_drk.fits',"
    print "        thebasedark='basedark_drk.fits')\n"
    
    print 'Be sure to set the header keyword PCTECORR to COMPLETE in super-darks to use them with stis_cti:'
    print "    fits.setval('weekdark_drk.fits', 'PCTECORR', value='COMPLETE')\n"
    
    print 'Point each science file to the new super-dark:'
    print "    fits.setval('science/filename_raw.fits', 'DARKFILE', value='stisref$weekdark_drk.fits')\n"
    
    print 'Also, don\'t run stis_cti with the --crds_update option, as this will override your custom'
    print 'super-dark.'



#!/usr/bin/env python

from nose.tools import assert_equals, assert_false, assert_raises, assert_not_equals
import os
import copy
import datetime
from astropy.io import fits

from stis_cti import *
from archive_dark_query import archive_dark_query

# ----------------------------------------------------------------------------------------
# These functions are used to write test files used by the unit tests

def write_file(test_file):
    # Create and write out a test FITS file:
    if not os.access(os.path.curdir, os.W_OK):
        raise Exception('Can\'t write test file to CWD!')
    
    hdu = fits.PrimaryHDU()
    
    hdu.header['TELESCOP'] = ('HST',           'telescope used to acquire data')
    hdu.header['INSTRUME'] = ('STIS',          'identifier for instrument used to acquire data')
    hdu.header['TDATEOBS'] = ('2012-07-14',    'UT date of start of first exposure in file')
    hdu.header['TTIMEOBS'] = ('04:35:52',      'UT start time of first exposure in file')
    hdu.header['OBSTYPE']  = ('SPECTROSCOPIC', 'observation type - imaging or spectroscopic')
    hdu.header['OBSMODE']  = ('ACCUM',         'operating mode')
    hdu.header['SUBARRAY'] = (False,           'data from a subarray (T) or full frame (F)')
    hdu.header['DETECTOR'] = ('CCD',           'detector in use: NUV-MAMA, FUV-MAMA, or CCD')
    hdu.header['CCDAMP']   = ('D',             'CCD amplifier read out (A,B,C,D)')
    hdu.header['CCDGAIN']  = (1,               'commanded gain of CCD')
    hdu.header['CCDOFFST'] = (3,               'commanded CCD bias offset')
    hdu.header['BINAXIS1'] = (1,               'axis1 data bin size in unbinned detector pixels')
    hdu.header['BINAXIS2'] = (1,               'axis2 data bin size in unbinned detector pixels')
    
    hdu.writeto(test_file, output_verify='exception', clobber=True)


def write_pctetab(pctetab):
    if not os.access(os.path.curdir, os.W_OK):
        raise Exception('Can\'t write test PCTETAB file to CWD!')
    
    hdu = fits.PrimaryHDU()
    
    hdu.header['FILENAME'] = pctetab
    hdu.header['FILETYPE'] = 'PIXCTE'                                                            
    hdu.header['TELESCOP'] = 'HST'                                                            
    hdu.header['USEAFTER'] = 'Oct 01 1996 00:00:00'                                                
    hdu.header['PEDIGREE'] = 'INFLIGHT 01/10/1996 25/06/2012'                                      
    hdu.header['DESCRIP']  = 'Parameters needed for pixel-based CTE correction ------------------' 
    hdu.header['NCHGLEAK'] = (1, 'number of chg_leak extensions')
    hdu.header['INSTRUME'] = 'STIS'                                                            
    hdu.header['DETECTOR'] = 'CCD'                                                            
    hdu.header['SIM_NIT']  = (7, 'number of readout simulations done per column')
    hdu.header['SHFT_NIT'] = (4, 'the number of shifts each column readout simula')
    hdu.header['RN_CLIP']  = (5.6, 'Read noise level in electrons.')
    hdu.header['NSEMODEL'] = (1, 'Read noise smoothing algorithm.')
    hdu.header['SUBTHRSH'] = (-30.0, 'Over-subtraction correction threshold.')
    hdu.header['PCTE_VER'] = ('0.1_alpha', 'Version of PCTETAB')
    
    hdu.writeto(pctetab, output_verify='exception', clobber=True)


def write_superdark(superdark):
    if not os.access(os.path.curdir, os.W_OK):
        raise Exception('Can\'t write test superdark file to CWD!')
    
    hdu = fits.PrimaryHDU()
    
    hdu.header['FILENAME'] = superdark
    hdu.header['FILETYPE'] = 'DARK IMAGE'
    hdu.header['TELESCOP'] = 'HST'
    hdu.header['REF_TEMP'] = 18
    hdu.header['DRK_VS_T'] = 0.07
    hdu.header['PCTECORR'] = 'COMPLETE'
    hdu.header['PCTETAB']  = 'pctetab$a01_stis_pcte.fits'
    hdu.header['USEAFTER'] = 'Jul 24 2012 00:59:27'
    hdu.header['PEDIGREE'] = 'INFLIGHT 01/10/1996 25/06/2012'
    hdu.header['DESCRIP']  = 'Weekly gain=1 dark for STIS CCD data taken after Jul 24 2012-------'
    hdu.header['INSTRUME'] = 'STIS'
    hdu.header['DETECTOR'] = ('CCD', 'detector in use: CCD')
    
    hdu.header['PCTEFRAC'] = (1.192887185841836,              'CTE time scaling value')
    hdu.header['PCTESMIT'] = (7,                              'PCTE readout simulation iterations')
    hdu.header['PCTESHFT'] = (4,                              'PCTE readout number of shifts')
    hdu.header['PCTERNCL'] = (5.6,                            'PCTE readnoise amplitude')
    hdu.header['PCTENSMD'] = (1,                              'PCTE readnoise mitigation algorithm')
    hdu.header['PCTETRSH'] = (-30.0,                          'PCTE over-subtraction threshold')
    hdu.header['PCTE_VER'] = ('0.1_alpha',                    'Version of PCTETAB')
    hdu.header['CTE_NAME'] = ('PixelCTE 2012',                'name of CTE algorithm')
    hdu.header['CTE_VER']  = ('3.2',                          'version of CTE algorithm')
    hdu.header['BASEDARK'] = ('$ctitest/bdark_d133_drk.fits', 'Used to make weekdark')
    
    hdu.header['HISTORY'] = 'blah, blah blah'
    hdu.header['HISTORY'] = 'The following input files were used:'
    hdu.header['HISTORY'] = 'abc_cte.fits'
    hdu.header['HISTORY'] = 'def_cte.fits'
    hdu.header['HISTORY'] = 'hij_cte.fits'
    hdu.header['HISTORY'] = ''
    hdu.header['HISTORY'] = 'blah2, blah2, blah2'
    
    hdu.writeto(superdark, output_verify='exception', clobber=True)


# ----------------------------------------------------------------------------------------
# Here are the unit tests:

class TestPaths(object):
    '''
    Tests functionality of stis_cti.resolve_iraf_file
    '''
    @classmethod
    def setup_class(cls):
        os.environ['toref'] = '/grp/hst/cdbs/oref/'
        # Undefine undoref if it exists:
        if os.environ.has_key('undoref'):
            os.environ.pop('undoref')
    
    @classmethod
    def teardown_class(cls):
        os.environ.pop('toref')
    
    def test_no_dollar(self):
        assert_equals(resolve_iraf_file('filename.fits'), 'filename.fits')
        assert_equals(resolve_iraf_file('/dir/filename.fits'), '/dir/filename.fits')
        assert_equals(resolve_iraf_file('dir/filename.fits'), 'dir/filename.fits')
    
    def test_dollar_at_beginning(self):
        assert_equals(resolve_iraf_file('$toref/filename.fits'), '/grp/hst/cdbs/oref/filename.fits')
        assert_equals(resolve_iraf_file('$toref/path/filename.fits'), '/grp/hst/cdbs/oref/path/filename.fits')
    
    def test_dollar_in_middle(self):
        assert_equals(resolve_iraf_file('toref$filename.fits'), '/grp/hst/cdbs/oref/filename.fits')
        assert_equals(resolve_iraf_file('toref$path/filename.fits'), '/grp/hst/cdbs/oref/path/filename.fits')
    
    def test_environ_var_undefined(self):
        with assert_raises(IOError) as cm:
            resolve_iraf_file('undoref$filename.fits')
        ex = cm.exception
        assert_equals(ex.message, "Can't resolve environmental variable 'undoref'.")


class TestFileFiltering(object):
    '''
    Tests functionality of stis_cti.viable_ccd_file
    '''
    test_file = 'testfits_2334134234_raw.fits'
    
    def setup(self):
        write_file(self.test_file)
    
    def teardown(self):
        os.remove(self.test_file)
    
    def test_viable_file(self):
        assert viable_ccd_file(self.test_file)
    
    def test_viable_file_lenient(self):
        assert viable_ccd_file(self.test_file, \
            earliest_date_allowed = datetime.datetime(1990,1,1,0,0,0), \
            amplifiers_allowed = ['A','B','C','D'], \
            gains_allowed = [1,2,4,8], \
            offsts_allowed = range(9))
    
    def test_acq_file_reject(self):
        fits.setval(self.test_file, 'OBSMODE', value='ACQ')
        assert_false(viable_ccd_file(self.test_file))
    
    def test_gain3_file_reject(self):
        fits.setval(self.test_file, 'CCDGAIN', value=3)
        assert_false(viable_ccd_file(self.test_file))
    
    def test_gain4_file(self):
        fits.setval(self.test_file, 'CCDGAIN', value=4)
        assert viable_ccd_file(self.test_file)

    def test_pre_sm4_reject(self):
        fits.setval(self.test_file, 'TDATEOBS', value='1990-01-01')
        assert_false(viable_ccd_file(self.test_file))

    def test_binned_data_reject1(self):
        fits.setval(self.test_file, 'BINAXIS1', value=2)
        assert_false(viable_ccd_file(self.test_file))
    
    def test_binned_data_reject2(self):
        fits.setval(self.test_file, 'BINAXIS2', value=2)
        assert_false(viable_ccd_file(self.test_file))
    
    def test_subarray_reject(self):
        fits.setval(self.test_file, 'SUBARRAY', value=True)
        assert_false(viable_ccd_file(self.test_file))


class Test_determine_input_science(object):
    '''
    Tests functionality of stis_cti.determine_input_science
    '''
    test_dir   = 'dir_2334134234'
    test_file  = os.path.join(test_dir, 'testfits_001_raw.fits')
    test_file2 = os.path.join(test_dir, 'testfits_002_raw.fits')
    
    def setup(self):
        if not os.access(os.path.curdir, os.W_OK):
            raise Exception('Can\'t write test dir and file to CWD!')
        
        # Setup directory structure:
        if os.path.exists(self.test_file):
            os.remove(self.test_file)
        if os.path.exists(self.test_file2):
            os.remove(self.test_file2)
        if os.path.exists(self.test_dir):
            os.rmdir(self.test_dir)
        os.mkdir(self.test_dir)
        
        # Write FITS file:
        write_file(self.test_file)
    
    def teardown(self):
        os.remove(self.test_file)
        if os.path.exists(self.test_file2):
            os.remove(self.test_file2)
        os.rmdir(self.test_dir)
    
    def test_file_written(self):
        assert os.path.exists(self.test_file)
    
    def test_filter_pass(self):
        assert_equals(determine_input_science(self.test_dir, False, False), [self.test_file])
    
    def test_filter_date_reject(self):
        fits.setval(self.test_file, 'TDATEOBS', value='1992-01-01')
        assert_raises(FileError, determine_input_science, self.test_dir, False, False)
    
    def test_filter_date_allow_pass(self):
        fits.setval(self.test_file, 'TDATEOBS', value='1992-01-01')
        assert determine_input_science(self.test_dir, True, False)
    
    def test_filter_date_partial_reject(self):
        write_file(self.test_file2)
        fits.setval(self.test_file2, 'TDATEOBS', value='1992-01-01')
        assert_equals(determine_input_science(self.test_dir, False, False), [self.test_file])


class Test_check_pctetab_version(object):
    '''
    Tests functionality of stis_cti.check_pctetab_version
    '''
    pctetab = 'a00_stis_000test37155_pcte.fits'
    
    def setup(self):
        write_pctetab(self.pctetab)
    
    def teardown(self):
        os.remove(self.pctetab)
     
    def test_pctetab_ver_pass(self):
        assert check_pctetab_version(self.pctetab, False, '0.1', '1.999')
    
    def test_pctetab_ver_reject(self):
        assert_raises(VersionError, check_pctetab_version, self.pctetab, False, '1.0', '1.999')
    
    def test_pctetab_ver_file_reject(self):
        fits.setval(self.pctetab, 'PCTE_VER', value='3.0_beta')
        assert_raises(VersionError, check_pctetab_version, self.pctetab, False, '0.1', '1.999')


class Test_superdark_hash(object):
    '''
    Tests functionality of stis_cti.superdark_hash
    '''
    pctetab = 'a00_stis_000test37166_pcte.fits'
    superdark = 'd001_testfile57254_drk.fits'
    
    def setup(self):
        write_superdark(self.superdark)
        write_pctetab(self.pctetab)
    
    def teardown(self):
        os.remove(self.superdark)
        os.remove(self.pctetab)
    
    def test_manual_inputs_pass(self):
        assert_equals(
            superdark_hash(sim_nit=7, shft_nit=4, rn_clip=5.6, nsemodel=1, subthrsh=-30.0, 
                           pcte_ver='0.1_alpha', files=[]), 
            superdark_hash(pctetab=self.pctetab, files=[]))
    
    def test_manual_inputs2_pass(self):
       # superdark_hash should ignore text after '_' in pcte_ver
        assert_equals(
            superdark_hash(sim_nit=7, shft_nit=4, rn_clip=5.6, nsemodel=1, subthrsh=-30.0, 
                           pcte_ver='0.1_beta', files=[]), 
            superdark_hash(pctetab=self.pctetab, files=[]))
    
    def test_manual_inputs3_pass(self):
        # try with a file list specified
        assert_equals(
            superdark_hash(sim_nit=7, shft_nit=4, rn_clip=5.6, nsemodel=1, subthrsh=-30.0, 
                           pcte_ver='0.1_alpha', files=['abc_cte.fits', 'def_cte.fits', 'hij_cte.fits']), 
            superdark_hash(pctetab=self.pctetab, files=['abc_cte.fits', 'def_cte.fits', 'hij_cte.fits']))
    
    def test_manual_inputs4_pass(self):
        assert_equals(
            superdark_hash(sim_nit=7, shft_nit=4, rn_clip=5.6, nsemodel=1, subthrsh=-30.0, 
                           pcte_ver='0.1_alpha', files=['abc_cte.fits', 'def_cte.fits', 'hij_cte.fits']), 
            superdark_hash(superdark=self.superdark))
    
    def test_manual_inputs_reject(self):
        # changed file list
        assert_not_equals(
            superdark_hash(sim_nit=7, shft_nit=4, rn_clip=5.6, nsemodel=1, subthrsh=-30.0, 
                           pcte_ver='0.1_alpha', files=[]), 
            superdark_hash(pctetab=self.pctetab, files=['otherfile_cte.fits']))
    
    def test_manual_inputs2_reject(self):
        # changed sim_nit
        assert_not_equals(
            superdark_hash(sim_nit=70, shft_nit=4, rn_clip=5.6, nsemodel=1, subthrsh=-30.0, 
                           pcte_ver='0.1_alpha', files=[]), 
            superdark_hash(pctetab=self.pctetab, files=[]))
    
    def test_manual_inputs3_reject(self):
        # changed pcte_ver
        assert_not_equals(
            superdark_hash(sim_nit=7, shft_nit=4, rn_clip=5.6, nsemodel=1, subthrsh=-30.0, 
                           pcte_ver='0.2_alpha', files=[]), 
            superdark_hash(pctetab=self.pctetab, files=[]))
    
    def test_superdark_pass(self):
        assert_equals(
            superdark_hash(superdark=self.superdark, files=[]), 
            superdark_hash(pctetab=self.pctetab, files=[]))
    
    def test_superdark2_pass(self):
        assert_equals(
            superdark_hash(superdark=self.superdark), 
            superdark_hash(pctetab=self.pctetab, files=['abc_cte.fits', 'def_cte.fits', 'hij_cte.fits']))
    
    def test_superdark3_pass(self):
        fits.setval(self.superdark, 'PCTE_VER', value='0.1_gamma')
        assert_equals(
            superdark_hash(superdark=self.superdark), 
            superdark_hash(pctetab=self.pctetab, files=['abc_cte.fits', 'def_cte.fits', 'hij_cte.fits']))
    
    def test_superdark_reject(self):
        assert_not_equals(
            superdark_hash(superdark=self.superdark), 
            superdark_hash(pctetab=self.pctetab, files=[]))
    
    def test_superdark2_reject(self):
        fits.setval(self.superdark, 'PCTENSMD', value=5)
        assert_not_equals(
            superdark_hash(superdark=self.superdark), 
            superdark_hash(pctetab=self.pctetab, files=['abc_cte.fits', 'def_cte.fits', 'hij_cte.fits']))
    
    def test_superdark3_reject(self):
        fits.setval(self.pctetab, 'nsemodel', value=6)
        assert_not_equals(
            superdark_hash(superdark=self.superdark), 
            superdark_hash(pctetab=self.pctetab, files=['abc_cte.fits', 'def_cte.fits', 'hij_cte.fits']))
    
    def test_superdark4_reject(self):
        fits.setval(self.superdark, 'PCTE_VER', value='0.2_alpha')
        assert_not_equals(
            superdark_hash(superdark=self.superdark), 
            superdark_hash(pctetab=self.pctetab, files=['abc_cte.fits', 'def_cte.fits', 'hij_cte.fits']))
    
    def test_superdark4_reject(self):
        fits.setval(self.pctetab, 'PCTE_VER', value='0.2_alpha')
        assert_not_equals(
            superdark_hash(superdark=self.superdark), 
            superdark_hash(pctetab=self.pctetab, files=['abc_cte.fits', 'def_cte.fits', 'hij_cte.fits']))
    

class Test_archive_dark_query(object):
    '''
    Tests functionality of stis_cti.archive_dark_query
    '''
    test_file = 'testfits_2334134667_raw.fits'
    
    @classmethod
    def setup_class(cls):
        write_file(cls.test_file)
        cls.anneal = archive_dark_query([cls.test_file], None, None, False, False)
    
    @classmethod
    def teardown_class(cls):
        os.remove(cls.test_file)
    
    def test_number_of_anneals(self):
        assert_equals(len(self.anneal), 1)
    
    def test_anneal_index(self):
        assert_equals(self.anneal[0]['index'], 133)
    
    def test_anneal_darks(self):
        darks = [d['exposure'] for d in self.anneal[0]['darks']]
        assert_equals(set(darks), \
            set(['OBVM3XH9Q', 'OBVM3YHHQ', 'OBVM3ZN5Q', 'OBVM40NCQ', 'OBVM41T2Q', 'OBVM42TFQ', 
                 'OBVM43YEQ', 'OBVM44YOQ', 'OBVM45FQQ', 'OBVM46G7Q', 'OBVM47LNQ', 'OBVM48MCQ', 
                 'OBVM49ANQ', 'OBVM4AAVQ', 'OBVM4BH5Q', 'OBVM4CHJQ', 'OBVM4DLRQ', 'OBVM4EM2Q', 
                 'OBVM4FRJQ', 'OBVM4GRQQ', 'OBVM4HYVQ', 'OBVM4IZ5Q', 'OBVM4JG4Q', 'OBVM4KGCQ', 
                 'OBVM4LO1S', 'OBVM4MOIS', 'OBVM4NABQ', 'OBVM4OAMQ', 'OBVM4PI2Q', 'OBVM4QIAQ', 
                 'OBVM4RPDQ', 'OBVM4SPJQ', 'OBVM4TW9Q', 'OBVM4UWDQ', 'OBVM4VCOQ', 'OBVM4WCXQ', 
                 'OBVM4XJ6Q', 'OBVM4YJCQ', 'OBVM4ZOMQ', 'OBVM50POQ', 'OBVM51A4Q', 'OBVM52ABQ', 
                 'OBVM53FIQ', 'OBVM54FNQ', 'OBVM55LTQ', 'OBVM56M5Q', 'OBVM57T3S', 'OBVM58T8S', 
                 'OBVM59W6S', 'OBVM5AWDS', 'OBVM5BBQQ', 'OBVM5CBUQ', 'OBVM5DG7Q', 'OBVM5EGBQ', 
                 'OBVM5FALQ', 'OBVM5GAUQ', 'OBVM5HFTQ', 'OBVM5IFPQ', 'OBVM5JKLQ', 'OBVM5KLPQ']))
    
    def test_undefined_fits_file(self):
        undefined_filename = 'testfits_undefined_28739723_raw.fits'
        assert_raises(IOError, archive_dark_query, [undefined_filename], None, None, False, False)


class Test_check_for_old_output_files(object):
    '''
    Tests functionality of stis_cti.check_for_old_output_files
    '''
    test_dir = 'dir_8364834236'
    rootname = 'testfits67733'
    
    test_files_leave  = [rootname + '_raw.fits', rootname + '_flt.fits']
    test_files_remove = [rootname + '_s2c.fits', rootname + '_flc.fits',
                         rootname + '_blt.fits', rootname + '_cte_flt.fits']
    # Prepend these files with test_dir:
    test_files_leave  = [os.path.join(test_dir, f) for f in test_files_leave]
    test_files_remove = [os.path.join(test_dir, f) for f in test_files_remove]
    
    # Combine both of these arrays into one:
    test_files = copy.deepcopy(test_files_leave)
    test_files.extend(test_files_remove)
    
    output_mapping = {
        'cte_flt.fits' : 'flc.fits' ,
        'cte_crj.fits' : 'crc.fits' ,
        'cte_sx2.fits' : 's2c.fits' ,
        'cte_x2d.fits' : 'x2c.fits' ,
        'cte_sx1.fits' : 's1c.fits' ,
        'cte_x1d.fits' : 'x1c.fits' ,
        'blt_tra.txt'  : 'trb.txt'  ,
        'cte_tra.txt'  : 'trc.txt'  ,
        'blt.fits'     : '<pass>'   ,
        'cte.fits'     : '<pass>'   }
    
    def setup(self):
        if os.path.exists(self.test_dir):
            raise IOError('test_dir already exists: {}'.format(self.test_dir))
        else:
            os.mkdir(self.test_dir)
    
    def teardown(self):
        for file in self.test_files:
            if os.path.exists(file):
                os.remove(file)
        os.rmdir(self.test_dir)
    
    def test_only_good_files(self):
        for file in self.test_files_leave:
            write_file(file)
        assert check_for_old_output_files([self.rootname], self.test_dir, 
            self.output_mapping, False, False)
    
    def test_files_already_exist_exception(self):
        for file in self.test_files:
            write_file(file)
        assert_raises(IOError, check_for_old_output_files, [self.rootname], self.test_dir, 
            self.output_mapping, False, False)
    
    def test_files_already_exist_exception_none_left(self):
        for file in self.test_files_remove:
            write_file(file)
        assert_raises(IOError, check_for_old_output_files, [self.rootname], self.test_dir, 
            self.output_mapping, False, False)
    
    def test_files_already_exist_clean(self):
        for file in self.test_files:
            write_file(file)
        assert check_for_old_output_files([self.rootname], self.test_dir, 
            self.output_mapping, True, False)
    
    def test_only_good_files_clean(self):
        for file in self.test_files_leave:
            write_file(file)
        assert check_for_old_output_files([self.rootname], self.test_dir, 
            self.output_mapping, True, False)
    
    def test_only_bad_files_clean(self):
        for file in self.test_files_remove:
            write_file(file)
        assert check_for_old_output_files([self.rootname], self.test_dir, 
            self.output_mapping, True, False)


# ----------------------------------------------------------------------------------------
if __name__ == '__main__':
    import nose
    nose.runmodule(argv=['nose', '--verbosity=2'])



from astropy.io import fits
import glob
from collections import Counter
import datetime
import urllib, urllib2
from numpy import size, shape

__author__  = 'Sean Lockwood'
__version__ = '0.1.2'


# Data container for our dark exposure search results:
class darkType(dict):
    def __str__(self):
        try:
            return self['exposure']
        except KeyError:
            return 'undefined'
    def __repr__(self):
        return '<' + self.__str__() + '>'


class annealType(dict):
    def __str__(self):
        try:
            return '<' + '{:03d}'.format(self['index']) + '; ' + \
                   self['start'].isoformat(' ') + ' - ' + self['end'].isoformat(' ') + '>'
        except KeyError:
            return '<undefined>'
    def __repr__(self):
        return self.__str__()


def get_proposal_ids(abstract='stis, +ccd', title='dark, +monitor'):
    '''
    Perform an HST Abstract search for STIS/CCD Dark Monitor programs.
    '''
    
    url = 'http://archive.stsci.edu/hst/abstract.html'
    form_data = urllib.urlencode({
        'abstract': abstract,      # String to be searched for within the abstract
        'atitle':   title,         # String to be searched for within the title
        'checkbox': 'no',          # Display abstract?
        'submit':   'submit'})
    
    # Submit POST request:
    try:
        response = urllib2.urlopen(url, form_data)  # Needs better error handling!
        text = response.read().split('\n')
        response.close()
    except (urllib2.URLError, urllib2.HTTPError) as e:
        print 'Please check your internet connection!'
        raise e
    
    # Parse returned text for proposal IDs:
    text = filter(lambda x: 'proposal_search.php' in x, text)
    proposals = map(lambda y: int(y.split('>')[1].split('<')[0]), text)
    
    return proposals


def read_dark_exposures():
    '''
    Query the HST archive for STIS/CCD dark exposures.
    '''
    
    # URL for HTTP GET request to the HST archive:
    url = 'http://archive.stsci.edu/hst/search.php?' + urllib.urlencode([
        ('sci_instrume'         , 'STIS'                                         ),
        ('sci_instrument_config', 'STIS/CCD'                                     ),
        ('sci_targname'         , 'DARK'                                         ),
        ('sci_aec'              , 'C'                                            ),
        ('resolve'              , 'don%27tresolve'                               ),
        ('max_records'          , '50000'                                        ),
        ('selectedColumnsCsv'   , 'sci_data_set_name,sci_pep_id,sci_start_time,sci_actual_duration'),
        ('ordercolumn1'         , 'sci_start_time'                               ),
        ('showquery'            , 'off'                                          ),
        ('skipformat'           , 'on'                                           ),
        ('nonull'               , 'on'                                           ),
        ('outputformat'         , 'CSV'                                          ),
        ('action'               , 'Search'                                       )])
    
    # Also interested in:
    #    stis_ref_data..ssr_ccdgain
    #    stis_ref_data..ssr_ccdamp
    #    stis_ref_data..ssr_ccdoffst
    
    # Submit HTTP GET request:
    try:
        response = urllib2.urlopen(url)
        data = response.readlines()
        response.close()
    except (urllib2.URLError, urllib2.HTTPError) as e:
        print 'Please check your internet connection!'
        raise e
    
    darks = []
    for line in data:
        if line != '\n' and 'Dataset' not in line and 'string' not in line:
            exposure, proposid, time, exptime = line.split(',')
            exptime = float(exptime.split('\n')[0])
            proposid = int(proposid)
            time = datetime.datetime.strptime(time, '%Y-%m-%d %H:%M:%S')
            darks.append(darkType(exposure=exposure, proposid=proposid, datetime=time, exptime=exptime))
    
    return darks


def get_anneal_boundaries(delta_days=5, min_exptime=None, verbose=False):
    '''
    Determines the annealing period boundaries and all associated darks within.
    '''
    
    if min_exptime is None:
        min_exptime = 960.0  # seconds
    else:
        min_exptime = float(min_exptime)
    
    if verbose:
        print 'Minimum exposure time = ', min_exptime, ' s'
        print
    
    print 'Querying MAST archive for dark and anneal program IDs...'
    dark_programs = get_proposal_ids()
    dark_programs.extend([7092, 7601, 7802, 7926, 7948, 7949])  # Older programs that didn't match this search pattern
    dark_programs = list(set(dark_programs))
    dark_programs.sort()
    
    anneal_programs = get_proposal_ids(abstract='', title='STIS CCD Hot Pixel Annealing')
    anneal_programs.sort()
    
    if verbose:
        print
        print 'Dark proposal IDs:       '
        print ', '.join(map(lambda x: str(x), dark_programs))
        print
        print 'Annealing proposal IDs:  '
        print ', '.join(map(lambda x: str(x), anneal_programs))
        print
    
    print 'Querying MAST archive for darks...' 
    darks = read_dark_exposures()
    
    print 'Parsing archive results...'
    print
    
    anneal_exposures = filter(lambda x: x['proposid'] in anneal_programs, darks)
    
    # Find delta time between neighboring anneal exposures:
    anneal_start_boundary = [anneal_exposures[0]['datetime'] - datetime.timedelta(days=60)]
    anneal_end_boundary = [anneal_exposures[0]['datetime']]  # Or, use a later exposure < delta_days?
    
    previous_anneal = anneal_exposures[0]
    for anneal_exposure in anneal_exposures[1:]:
        delta = anneal_exposure['datetime'] - previous_anneal['datetime']
        if (delta.total_seconds() / 3600. / 24.) >= delta_days:
            anneal_start_boundary.append(previous_anneal['datetime'])
            anneal_end_boundary.append(anneal_exposure['datetime'])  # Or, use start time to avoid gaps?
            previous_anneal = anneal_exposure
    
    # Add an additional annealing period after the last anneal (end date is approximate):
    anneal_start_boundary.append(anneal_exposure['datetime'])
    anneal_end_boundary.append(anneal_exposure['datetime'] + datetime.timedelta(days=60))
    
    # Determine which darks fall within each annealing period:
    anneals = []
    
    for i, (a, b) in enumerate(zip(anneal_start_boundary, anneal_end_boundary)):
        darks_within_anneal = filter( \
            lambda x: x['datetime'] >= a and 
                      x['datetime'] < b  and
                      x['proposid'] in dark_programs and
                      x['exptime'] >= min_exptime,
            darks)
        
        anneals.append(annealType(index=i, start=a, end=b, darks=list(darks_within_anneal)))
    
    if (verbose >= 2):
        # Print information about all annealing periods:
        print "Dark data found for all annealing periods:"
        for i, a in enumerate(anneals):
            print ('{i:3d} {start:} - {end:} ({length:4d} days), {total_exposures:3d} total: ' + \
                   '{exptimes} {proposals}').format( 
                i = i, 
                start = a['start'].isoformat(' '), 
                end = a['end'].isoformat(' '), 
                length = (a['end'] - a['start']).days, 
                total_exposures = shape(a['darks'])[0], 
                exptimes = Counter(map(lambda x: int(x['exptime']), a['darks'])), 
                proposals = list(set(map(lambda x: x['proposid'], a['darks']))))
        print
    
    return anneals


def darks_url(exposures):
    # Build a URL to retrieve matched datasets from MAST:
    url = 'http://archive.stsci.edu/hst/search.php?' + urllib.urlencode([
        ('sci_instrume'         , 'STIS'              ),
        ('sci_instrument_config', 'STIS/CCD'          ),
        ('sci_targname'         , 'DARK'              ),
        ('sci_aec'              , 'C'                 ),
        ('resolve'              , 'don\'tresolve'     ),
        ('sci_data_set_name'    , ','.join(exposures) ),
        ('max_records'          , '50000'             ),
        ('max_rpp'              , '5000'              ),
        ('ordercolumn1'         , 'sci_start_time'    ),
        ('action'               , 'Search'            )])
    
    return url


def archive_dark_query(files, anneal_data=None, min_exptime=None, verbose=False, print_url=True):
    '''
    Queries the MAST archive to determine which component darks are needed to 
    generate a CTI-corrected super-dark.
    '''
    
    # Convert to a list of filenames if only a string is provided:
    if isinstance(files, str):
        files = [files]
    
    # Apply glob to every element of files to evaluate search patterns:
    files_parsed = []
    for f in files:
        files_parsed.extend(glob.glob(f))
    files = files_parsed
    
    if verbose:
        print 'Input files:'
        for f in files:
            print f
        print
    
    if len(files) == 0:
        raise IOError('No files specified/found.')
    
    # Determine time boundaries for each annealing period:
    # (This is costly, so pass the value into this function when calling multiple times.)
    if anneal_data == None:
        anneal_data = get_anneal_boundaries(verbose=verbose, min_exptime=min_exptime)
    else:
       if verbose:
           print 'Skipping re-population of anneal_data.'
           print
    
    # Populate matched exposures uniquely:
    matches = list()
    
    for file in files:
        with fits.open(file) as data:
            hdr0 = data[0].header
            date = hdr0['TDATEOBS']
            time = hdr0['TTIMEOBS']
            dt = datetime.datetime.strptime(date + ' ' + time, '%Y-%m-%d %H:%M:%S')
            match = filter(lambda x: dt >= x['start'] and dt < x['end'], anneal_data)[0]
            
            #matches.add(match)  # WON'T WORK WITH MUTABLE SUB-TYPES! ***
            matches.append(match)
            
            if verbose:
                print file
                print 'Annealing period:  ', match
                if verbose >= 2:
                    print ', '.join([str(i) for i in match['darks']])
                print
    
    # Remove duplicate entries based on the index field:
    matches = dict((item['index'], item) for item in matches).values()
    matches.sort(key=lambda x: x['index'])
    
    if verbose:
        print 'Unique annealing periods (' + str(shape(matches)[0]) + '):'
        print '\n'.join([str(i) for i in matches])
        print
    
    # Get the unique list of exposure names:
    all_exposures = set()
    for m in matches:
        for file in m['darks']:
            all_exposures.add(file['exposure'])
    all_exposures = list(all_exposures)
    all_exposures.sort()
    
    if verbose:
        print 'Unique exposures (' + str(shape(all_exposures)[0]) + '):'
        print ', '.join(all_exposures)
        print
    
    if print_url:
        url = darks_url(all_exposures)
        print 'Download darks via this link:'
        print
        print url
        print
    
    return matches



#!/usr/bin/env python

from setuptools import setup, Extension
import os
import numpy

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.rst')) as f:
    long_description = f.read()

StisPixCteCorr_module = Extension('stis_cti.StisPixCte_FixY',
    sources=['src/StisFixYCte.c', 'src/StisPixCteCorr_funcs.c', 'src/StisPixCte_FixY.c'],
    include_dirs=[numpy.get_include(), 'src/'])

setup(
    name = 'stis_cti',
    url = 'http://www.stsci.edu/instruments/stis/',
    version = '1.1',
    description = 'Pixel-based CTI-correction for HST/STIS CCD data',
    long_description = long_description,
    author = 'Sean Lockwood, Phil Hodge, Pey Lian Lim, W.J. Hack, J. Anderson, Matt Davis',
    author_email = 'help@stsci.edu',
    maintainer = 'Sean Lockwood',
    maintainer_email = 'lockwood@stsci.edu',
    license = 'BSD-new',
    keywords = ['astronomy'],
    classifiers = ['Programming Language :: Python',
                   'Programming Language :: C',
                   'Development Status :: 5 - Production/Stable',
                   'Operating System :: OS Independent',
                   'Intended Audience :: Science/Research',
                   'License :: OSI Approved :: BSD License',
                   'Topic :: Scientific/Engineering :: Astronomy',
                   'Topic :: Scientific/Engineering :: Physics',
                   'Topic :: Software Development :: Libraries :: Python Modules'],
    packages = ['stis_cti'],
    install_requires = ['setuptools', 'numpy', 'astropy>= 1.0.1', 'stistools>= 1.0.2', 
        'refstis>= 0.5.1', 'crds>= 1.3.0', 'stsci.tools>= 3.2.2'],
    scripts = ['scripts/stis_cti', 'scripts/archive_dark_query'],
    package_data = {'stis_cti': ['data/*.fits']},  # To do: Use entry_points instead
    ext_modules = [StisPixCteCorr_module],
    )



"""
Functions to apply pixel-based CTE correction to STIS images.

The algorithm implemented in this code was
described in detail by [Anderson]_ as available
online at:

http://adsabs.harvard.edu/abs/2010PASP..122.1035A

.. note::
    * This code only works for STIS/CCD but can be modified to work on other
      detectors.
    * It was developed for use with full-frame GAIN={1,4} FLT images as input.
    * It has not been fully tested with any other formats.
    * Noise is slightly enhanced in the output (see [Anderson]_).
    * This code assumes a linear time dependence for a given set of
      coefficients.
    * This algorithm does not account for traps with very long release timescale
      but it is not an issue.
    * This code also does not account for second-exposure effect.
    * Multi-threading support was not implemented in this version.

:Optional preprocessing for nonstandard FLT input:

If you are not using a fully calibrated FLT image as input,
you might also need to do one or more of the following before running the task:

    * Convert image to unit of electrons.

    * For combined image (e.g., superdark), set `noise_model` to 0.

    * Primary FITS header (EXT 0) must have these keywords populated:
        * ROOTNAME
        * INSTRUME (must be STIS)
        * DETECTOR (must be CCD)
        * NEXTEND (number of extensions = 3 * number of image sets)
        * CCDAMP (ABCD, AD, BC, A, B, C, or D)
        * TEXPSTRT
        * ATODGAIN

    * SCI FITS header must have these keywords populated:
        * EXTNAME (must be SCI)
        * EXTVER (1 or greater)

    * ERR FITS header must have these keywords populated:
        * EXTNAME (must be ERR)
        * EXTVER (1 or greater)

    * DQ FITS header must have these keywords populated:
        * EXTNAME (must be DQ)
        * EXTVER (1 or greater)

:Example:

To correct a set of STIS FLT images, with one new CTE-corrected image
for each input.

>>> from stistools import StisPixCteCorr
>>> StisPixCteCorr.CteCorr('o*q_flt.fits')

:References:

.. [Anderson] Anderson J. & Bedin, L.R., 2010, PASP, 122, 1035

"""

#### TAKEN OUT FROM FORMAL DOC BUT KEPT FOR HISTORY
#
#:Authors: Pey Lian Lim and W.J. Hack (Python), J. Anderson (Fortran), Matt Davis
#
#:Organization: Space Telescope Science Institute
#
#:History:
#    * 2010/09/01 PLL created this module.
#    * 2010/10/13 WH added/modified documentations.
#    * 2010/10/15 PLL fixed PCTEFILE lookup logic.
#    * 2010/10/26 WH added support for multiple file processing
#    * 2010/11/09 PLL modified `YCte`, `_PixCteParams` and `_DecomposeRN` to reflect noise improvement by JA. Also updated documentations.
#    * 2011/04/26 MRD Began updates for new CTE algorithm.
#    * 2011/07/20 MRD Updated to handle new PCTETAB containing time dependent
#      CTE characterization.
#    * 2011/12/08 MRD Updated with latest CTE algorithm 3.0.
#    * 2012/01/31 MRD Updated with latest CTE algorithm 3.1.
#    * 2012/05/11 PLL updated to version 3.2 to be consistent with CALACS.
#    * 2012/05/22 PLL removed {} formatting for Python 2.5/2.6 compatibility.
####################

# External modules
import os, shutil, time
import numpy
try:
    from astropy.io import fits as pyfits
except ImportError:
    import pyfits

try:
    from stsci.tools import teal
except:
    teal = None

from stsci.tools import parseinput

# Local modules
import StisPixCte_FixY as pcfy # C extension


__taskname__ = "StisPixCteCorr"
__version__ = "0.1"
__vdate__ = "2013-Mar-20"

# constants related to the CTE algorithm in use
STIS_CTE_NAME = 'PixelCTE 2012'
STIS_CTE_VER = '3.2'

# number of extensions per image set (SCI, ERR, DQ)
EXTN_PER_IMSET = 3


# general error for things related to this module
class StisPixCteError(Exception):
    pass


#--------------------------
def CteCorr(input, outFits='', read_noise=None, noise_model=None,
            oversub_thresh=None, sim_nit=None, shift_nit=None):
    """
    Run all the CTE corrections on all the input files.

    This function simply calls `YCte` on each input image
    parsed from the `input` parameter, and passes all remaining
    parameter values through unchanged.

    Parameters
    ----------
    input : str or list of str
        name of FLT image(s) to be corrected. The name(s) can be specified
        either as:

          * a single filename ('o1234567q_flt.fits')
          * a Python list of filenames
          * a partial filename with wildcards ('\*flt.fits')
          * filename of an ASN table ('o12345670_asn.fits')
          * an at-file (``@input``)

    outFits : str
        *USE DEFAULT IF `input` HAS MULTIPLE FILES.*
        CTE corrected image in the same
        directory as input. If not given, will use
        ROOTNAME_cte.fits instead. Existing file will
        be overwritten.

    read_noise : float, optional
        Read noise level.
        If None, takes value from PCTETAB header RN_CLIP keyword.

    noise_model : {0, 1, 2, None}, optional
        Noise mitigation algorithm.
        If None, takes value from PCTETAB header NSEMODEL keyword.

        0: No smoothing
        1: Normal smoothing
        2: Strong smoothing

    oversub_thresh : float, optional
        Pixels corrected below this value will be re-corrected.
        If None, takes value from PCTETAB header SUBTHRSH keyword.

    sim_nit : int, optional
        Number of times readout simulation is performed per column.
        If None, takes value from PCTETAB header SIM_NIT keyword.

    shift_nit : int, optional
        Number of times column is shifted during simulated readout.
        If None, takes value from PCTETAB header SHFT_NIT keyword.

    """
    # Parse input to get list of filenames to process
    infiles, output = parseinput.parseinput(input)

    # Process each file
    for file in infiles:
        YCte(file, outFits=outFits, read_noise=read_noise,
             noise_model=noise_model, oversub_thresh=oversub_thresh,
             sim_nit=sim_nit, shift_nit=shift_nit)


#--------------------------
def YCte(inFits, outFits='', read_noise=None, noise_model=None,
         oversub_thresh=None, sim_nit=None, shift_nit=None):
    """
    Apply correction for parallel CTE loss.

    Input image that is already de-striped is desired
    but not compulsory. Using image with striping
    will enhance the stripes in output. Calibrations
    that have been applied to FLT should not
    significantly affect the result.

    Notes
    -----
    * EXT 0 header will be updated. ERR arrays will be
      added in quadrature with 10% of the correction.
      DQ not changed.

    * Does not work on RAW but can be modified
      to do so.

    Parameters
    ----------
    inFits : str
        FLT image to be corrected.

    outFits : str
        CTE corrected image in the same
        directory as input. If not given, will use
        ROOTNAME_cte.fits instead. Existing file will
        be overwritten.

    read_noise, noise_model, oversub_thresh, sim_nit, shift_nit : see `CteCorr`

    Examples
    --------
    Correct a single FLT image and write output to 'o12345678_cte.fits':

    >>> import StisPixCteCorr
    >>> StisPixCteCorr.YCte('o12345678_flt.fits')

    """
    if noise_model not in (0, 1, 2, None):
        raise ValueError("noise_model must be one of (0, 1, 2, None).")

    # Start timer
    timeBeg = time.time()

    # For output files naming.
    # Store in same path as input.
    outPath = os.path.dirname( os.path.abspath(inFits) ) + os.sep
    rootname = pyfits.getval(inFits, 'ROOTNAME')
    print 'Performing pixel-based CTE correction on', rootname
    rootname = outPath + rootname

    # Construct output filename
    if not outFits:
        outFits = rootname + '_cte.fits'

    # Copy input to output
    shutil.copyfile(inFits, outFits)

    # Open output for correction
    pf_out = pyfits.open(outFits, mode='update')

    # For checking that the detector is supported
    detector = pf_out['PRIMARY'].header['DETECTOR']

    atodgain = pf_out['PRIMARY'].header.get('ATODGAIN', default=1.)
    
    ccdamp = pf_out['PRIMARY'].header.get('CCDAMP', default='D')
    if ccdamp != 'D':
        print 'Non-standard amplifer %s being corrected!' % ccdamp
        
    if ccdamp == 'A' or ccdamp == 'B':
        readout_dir = 1
    elif ccdamp == 'C' or ccdamp == 'D':
        readout_dir = -1
    else:
        raise ValueError("Amplifier %s is unrecognized!" % ccdamp)

    # For epoch-specific operations
    expstart = pf_out['PRIMARY'].header['TEXPSTRT']

    nextend = pf_out['PRIMARY'].header.get('NEXTEND', default=EXTN_PER_IMSET)
    nimsets = nextend // EXTN_PER_IMSET
    if nextend != nimsets * EXTN_PER_IMSET:
        raise ValueError('Number of extensions is %d; must be a multiple'
                         ' of %d.' % (nextend, EXTN_PER_IMSET))

    # This is just for STIS CCD.
    if detector != 'CCD':
        raise StisPixCteError('Invalid detector: StisPixCteCorr only'
                              ' supports STIS CCD.')

    # Read CTE params from file
    pctefile = pf_out['PRIMARY'].header['PCTETAB']
    pardict = _PixCteParams(pctefile, expstart)

    cte_frac = pardict['cte_frac']
    q_dtde = pardict['q_dtde']
    dtde_l = pardict['dtde_l']
    psi_node = pardict['psi_node']
    chg_leak = pardict['chg_leak']
    levels = pardict['levels']
    col_scale = pardict['col_scale']
    pcte_ver = pardict['pcte_ver']  # version of PCTETAB reference file

    if sim_nit is None:
        sim_nit = pardict['sim_nit']

    if shift_nit is None:
        shft_nit = pardict['shft_nit']
    else:
        shft_nit = shift_nit

    if read_noise is None:
        read_noise = pardict['read_noise']

    if noise_model is None:
        noise_model = pardict['noise_model']

    if oversub_thresh is None:
        oversub_thresh = pardict['oversub_thresh']

    # N in charge tail
    chg_leak_kt, chg_open_kt = pcfy.InterpolatePsi(chg_leak, psi_node)
    del chg_leak, psi_node

    # dtde_q: Marginal PHI at a given chg level.
    dtde_q = pcfy.InterpolatePhi(dtde_l, q_dtde, shft_nit)
    del dtde_l, q_dtde

    # finish interpolation along the Q dimension and reduce arrays to contain
    # only info at the levels specified in the levels array
    chg_leak_lt, chg_open_lt, dpde_l = \
      pcfy.FillLevelArrays(chg_leak_kt, chg_open_kt, dtde_q, levels)
    del chg_leak_kt, chg_open_kt, dtde_q

    ########################################
    # perform correction for each image set
    ########################################
    for imset in range(nimsets):
        # get data
        extn = 1 + imset * EXTN_PER_IMSET
        # flip the data so the readout amp is closest to row 0, column 0
        # (this assumes STIS amp D was used)
        # Modified (SL):  Does readout_dir==+1 work for 'A' and 'B'?
        scidata = pf_out[extn].data[::readout_dir,::readout_dir].copy().astype(numpy.float)
        errdata = pf_out[extn+1].data[::readout_dir,::readout_dir].copy().astype(numpy.float)

        # convert to electrons
        scidata *= atodgain
        errdata *= atodgain

        # separate signal and noise
        if noise_model in (1,2):
            sigdata, nsedata = pcfy.DecomposeRN(scidata, read_noise,
                                                noise_model)
        elif noise_model == 0:
            sigdata = scidata.copy()
            nsedata = numpy.zeros_like(sigdata)

        # set up cte frac array; flip amp D info from col_scale
        col_scale_temp = col_scale[ccdamp][:scidata.shape[1]][::readout_dir]
        cte_frac_arr = cte_frac * numpy.ones_like(scidata) * col_scale_temp
        cte_frac_arr = cte_frac_arr * \
            numpy.arange(1,scidata.shape[0]+1).reshape((scidata.shape[0],1))
        cte_frac_arr /= 1024.

        # call CTE correction routine. data must be in units of electrons.
        print 'Performing CTE correction for science extension %d.' % extn

        t1 = time.time()
        cordata = pcfy.FixYCte(sigdata, sim_nit, shft_nit, oversub_thresh,
                               cte_frac_arr, levels, dpde_l, chg_leak_lt,
                               chg_open_lt)
        t2 = time.time()

        print 'FixYCte took {:.3f} seconds for science extension {:d}.'.format(t2-t1, extn)

        # add noise back in
        findata = cordata + nsedata

        # add error as 10% of change; note that units are still electrons
        delta = 0.1 * numpy.abs(scidata - findata)
        errdata = numpy.sqrt(errdata**2 + delta**2)

        # convert corrected data back to DN
        findata /= atodgain
        errdata /= atodgain

        # copy corrected data back to image.
        pf_out[extn].data[:,:] = findata.astype(numpy.float32)[::readout_dir,::readout_dir]

        pf_out[extn+1].data[:,:] = errdata.astype(numpy.float32)[::readout_dir,::readout_dir]
        # end for imset in range(nimsets)

    # Update header
    pf_out['PRIMARY'].header['PCTECORR'] = 'COMPLETE'
    pf_out['PRIMARY'].header['PCTEFRAC'] = (cte_frac, 'CTE time scaling value')
    pf_out['PRIMARY'].header['PCTERNCL'] = (read_noise, 'PCTE readnoise amplitude')
    pf_out['PRIMARY'].header['PCTENSMD'] = (noise_model, 'PCTE readnoise mitigation algorithm')
    pf_out['PRIMARY'].header['PCTETRSH'] = (oversub_thresh, 'PCTE over-subtraction threshold')
    pf_out['PRIMARY'].header['PCTESMIT'] = (sim_nit, 'PCTE readout simulation iterations')
    pf_out['PRIMARY'].header['PCTESHFT'] = (shft_nit, 'PCTE readout number of shifts')
    pf_out['PRIMARY'].header['CTE_NAME'] = (STIS_CTE_NAME, 'name of CTE algorithm')
    pf_out['PRIMARY'].header['CTE_VER'] = (STIS_CTE_VER, 'version of CTE algorithm')
    pf_out['PRIMARY'].header['PCTE_VER'] = (pcte_ver, 'version of PCTETAB')
    pf_out['PRIMARY'].header.add_history('PCTECORR complete ...')
    
    pf_out['PRIMARY'].header['ATODGAIN'] = atodgain
    
    # Close output file
    pf_out.close()
    print outFits, 'written'

    # Stop timer
    timeEnd = time.time()
    print 'Run time:  {:.3f} secs'.format(timeEnd - timeBeg)


#--------------------------
def _PixCteParams(fitsTable, expstart):
    """
    Read params from PCTEFILE.

    .. note:: Environment variable pointing to
              reference file directory must exist.

    Parameters
    ----------
    fitsTable : str
        PCTEFILE from header.

    expstart : float
        MJD of exposure start time, TEXPSTRT in image header

    Returns
    -------
    A dictionary containing the following:

    cte_frac : float
        Time dependent CTE scaling.

    sim_nit : int
        Number of readout simulations to do for each column of data

    shft_nit : int
        Number of shifts to break each readout simulation into

    read_noise : float
        Maximum amplitude of read noise removed by DecomposeRN.

    noise_model : {0, 1, 2}
        Read noise smoothing algorithm selection.

    oversub_thresh : float
        CTE corrected pixels taken below this value are re-corrected.

    dtde_q : ndarray
        Charge levels at which dtde_l is parameterized

    dtde_l : ndarray
        PHI(Q).

    psi_node : ndarray
        N values for PSI(Q,N).

    chg_leak : ndarray
        PSI(Q,N).

    levels : ndarray
        Charge levels at which to do CTE evaluation

    col_scale : dict of ndarray
        Dictionary containing the column-by-column CTE scaling for each amp.

    """

    # Resolve path to PCTEFILE
    refFile = _ResolveRefFile(fitsTable)
    if not os.path.isfile(refFile):
        raise IOError, 'PCTEFILE not found: %s' % refFile

    # Open FITS table
    pf_ref = pyfits.open(refFile)

    # Read RN_CLIP value from header
    read_noise = pf_ref['PRIMARY'].header['RN_CLIP']

    # read NSEMODEL value
    noise_model = pf_ref['PRIMARY'].header['NSEMODEL']

    # read SIM_NIT value from header
    sim_nit = pf_ref['PRIMARY'].header['SIM_NIT']

    # read SHFT_NIT value from header
    shft_nit = pf_ref['PRIMARY'].header['SHFT_NIT']

    # read SUBTHRSH value from header
    oversub_thresh = pf_ref['PRIMARY'].header['SUBTHRSH']

    # read number of CHG_LEAK# extensions from the header
    nchg_leak = pf_ref['PRIMARY'].header['NCHGLEAK']

    # read dtde data from DTDE extension
    dtde_l = pf_ref['DTDE'].data['DTDE']
    q_dtde = pf_ref['DTDE'].data['Q']

    # read levels data from LEVELS extension
    levels = pf_ref['LEVELS'].data['LEVEL']

    # read scale data from CTE_SCALE extension
    scalemjd = pf_ref['CTE_SCALE'].data['MJD']
    scaleval = pf_ref['CTE_SCALE'].data['SCALE']

    cte_frac = _CalcCteFrac(expstart, scalemjd, scaleval)

    # there are nchg_leak CHG_LEAK# extensions. we need to find out which one
    # is the right one for our data.
    chg_leak_names = ['CHG_LEAK%i'%(i) for i in range(1,nchg_leak+1)]

    for n in chg_leak_names:
        mjd1 = pf_ref[n].header['MJD1']
        mjd2 = pf_ref[n].header['MJD2']

        if (expstart >= mjd1) and (expstart < mjd2):
            # read chg_leak data from CHG_LEAK extension
            psi_node = pf_ref[n].data['NODE']
            chg_leak = numpy.array(pf_ref[n].data.tolist(), dtype=numpy.float32)[:,1:]
            break

    # column-by-column CTE scaling
    col_scale = {}

    col_scale['A'] = pf_ref['COL_SCALE'].data['AMPA']
    col_scale['B'] = pf_ref['COL_SCALE'].data['AMPB']
    col_scale['C'] = pf_ref['COL_SCALE'].data['AMPC']
    col_scale['D'] = pf_ref['COL_SCALE'].data['AMPD']

    pcte_ver = pf_ref['PRIMARY'].header['PCTE_VER']

    # Close FITS table
    pf_ref.close()

    d = {}
    d['cte_frac'] = cte_frac
    d['sim_nit'] = sim_nit
    d['shft_nit'] = shft_nit
    d['oversub_thresh'] = oversub_thresh
    d['read_noise'] = read_noise
    d['noise_model'] = noise_model
    d['q_dtde'] = q_dtde
    d['dtde_l'] = dtde_l
    d['psi_node'] = psi_node
    d['chg_leak'] = chg_leak
    d['levels'] = levels
    d['col_scale'] = col_scale
    d['pcte_ver'] = pcte_ver

    return d


#--------------------------
def _ResolveRefFile(refText, sep='$'):
    """
    Resolve the full path to reference file.
    This could be replaced with existing STSDAS
    library function, if necessary.

    Assume standard syntax: dir$file.fits

    Parameters
    ----------
    refText : str
        The text to process.

    sep : char
        Separator between directory and file name.

    Returns
    -------
    f : str
        Full path to reference file.
    """

    s = refText.split(sep)
    n = len(s)
    if n > 1:
        p = os.getenv(s[0])
        if p:
            p += os.sep
        else:
            p = ''
        # End if
        f = p + s[1]
    else:
        f = os.path.abspath(refText)
    # End if
    return f


#--------------------------
def _CalcCteFrac(expstart, scalemjd, scaleval):
    """
    Calculate CTE_FRAC used to scale CTE according to time dependence.

    Parameters
    ----------
    expstart : float
        TEXPSTRT from header.

    scalemjd : ndarray
        MJD points for corresponding CTE scale values in scaleval

    scaleval : ndarray
        CTE scale values corresponding to MJDs in scalemjd

    Returns
    -------
    cte_frac : float
        Time scaling factor.
    """

    # Calculate CTE_FRAC
    cte_frac = pcfy.CalcCteFrac(expstart, scalemjd, scaleval)

    return cte_frac


#--------------------------
def _InterpolatePsi(chg_leak, psi_node):
    """
    Interpolates the `PSI(Q,N)` curve at all N from
    1 to 100.

    `PSI(Q,N)` models the CTE tail profile across N
    pixels from the original pixel for a given
    charge, Q. Up to 100 pixels are tracked. For
    post-SM4 STIS/CCD, CTE loss is within 100 pixels.
    Might be worse for WFPC2 since it is older and
    has faster readout time.

    .. note:: As this model is refined, future release
              might only have PSI(N) independent of Q.

    Parameters
    ----------
    chg_leak : ndarray
        PSI table data from PCTEFILE.

    psi_node : ndarray
        PSI node data from PCTEFILE.

    Returns
    -------
    chg_leak : ndarray
        Interpolated PSI.

    chg_open : ndarray
        Interpolated tail profile data.

    """

    chg_leak, chg_open = pcfy.InterpolatePsi(chg_leak, psi_node.astype(numpy.int32))

    return chg_leak, chg_open


#--------------------------
def _InterpolatePhi(dtde_l, q_dtde, shft_nit):
    """
    Interpolates the `PHI(Q)` at all Q from 1 to
    99999 (log scale).

    `PHI(Q)` models the amount of charge in CTE
    tail, i.e., probability of an electron being
    grabbed by a charge trap.

    Parameters
    ----------
    dtde_l : ndarray
        PHI data from PCTEFILE.

    q_dtde : ndarray
        Q levels at which dtde_l is defined, read from PCTEFILE

    shft_int : int
        Number of shifts performed reading out CCD

    Returns
    -------
    dtde_q : ndarray
        dtde_l interpolated at all PHI levels

    """

    dtde_q = pcfy.InterpolatePhi(dtde_l, q_dtde, shft_nit)

    return dtde_q


def _FillLevelArrays(chg_leak, chg_open, dtde_q, levels):
    """
    Interpolates CTE parameters to the charge levels specified in levels.

    Parameters
    ----------
    chg_leak : ndarray
        Interpolated chg_leak tail profile data returned by _InterpolatePsi.

    chg_open : ndarray
        Interpolated chg_open tail profile data returned by _InterpolatePsi.

    dtde_q : ndarray
        PHI data interpolated at all PHI levels as returned by
        _InterpolatePhi.

    levels : ndarray
        Charge levels at which output arrays will be interpolated.
        Read from PCTEFILE.

    Returns
    -------
    chg_leak_lt : ndarray
        chg_leak tail profile data interpolated at the specified charge levels.

    chg_open_lt : ndarray
        chg_open tail profile data interpolated at the specified charge levels.

    dpde_l : ndarray
        dtde_q interpolated and summed for the specified charge levels.

    tail_len : ndarray
        Array of maximum tail lengths for the specified charge levels.

    """

    chg_leak_lt, chg_open_lt, dpde_l = \
      pcfy.FillLevelArrays(chg_leak, chg_open, dtde_q, levels)

    return chg_leak_lt, chg_open_lt, dpde_l


#--------------------------
def _DecomposeRN(data_e, read_noise=4.25, noise_model=1):
    """
    Separate noise and signal.

    REAL DATA = SIGNAL + NOISE

    Parameters
    ----------
    data_e : ndarray
        SCI data in electrons.

    read_noise : float, optional
        Read noise level.

    noise_model : {0, 1, 2}, optional
        Noise mitigation algorithm.

        0: No smoothing
        1: Normal smoothing
        2: Strong smoothing

    Returns
    -------
    sigArr : ndarray
        Noiseless signal component in electrons.

    nseArr : ndarray
        Noise component in electrons.

    """
    if noise_model not in (0, 1, 2):
        raise ValueError("noise_model must be one of (0, 1, 2).")

    sigArr, nseArr = pcfy.DecomposeRN(data_e, read_noise, noise_model)

    return sigArr, nseArr


def _FixYCte(detector, cte_data, sim_nit, shft_nit, oversub_thresh,
             cte_frac, levels, dpde_l, chg_leak_lt, chg_open_lt):
    """
    Perform CTE correction on input data. It is best to perform some kind
    of readnoise smoothing on the data, otherwise the CTE algorithm will
    amplify the read noise. (In the read out process readnoise is added to
    the data after CTE blurring.)

    Parameters
    ----------
    detector : str
        DETECTOR from header.
        Currently only 'CCD' is supported.

    cte_data : ndarray
        Data in need of CTE correction. For proper results cte_data[0,x] should
        be next to the readout register and cte_data[-1,x] should be furthest.
        Data are processed a column at a time, e.g. cte_data[:,x] is corrected,
        then cte_data[:,x+1] and so on.

    sim_nit : int
        Number of readout simulation iterations to perform.

    shft_nit : int
        Number of readout shifts to do.

    oversub_thresh : float
        CTE corrected pixels taken below this value are re-corrected.

    cte_frac : ndarray
        CTE scaling image combining the time dependent and column-by-column
        CTE factors. Should be same shape as `cte_data`.

    levels : ndarray
        Levels at which CTE is evaluated as read from PCTEFILE.

    dpde_l : ndarray
        Parameterized amount of charge in CTE trails as a function of
        specific charge levels, as returned by _FillLevelArrays.

    chg_leak_lt : ndarray
        Tail profile data at charge levels specified by levels, as returned
        by _FillLevelArrays.

    chg_open_lt : ndarray
        Tail profile data at charge levels specified by levels, as returned
        by _FillLevelArrays.

    Returns
    -------
    corrected : ndarray
        Data after CTE correction algorithm applied. Same size and shape as
        input cte_data.

    """

    if detector == 'CCD':
        corrected = pcfy.FixYCte(cte_data, sim_nit, shft_nit, oversub_thresh,
                                 cte_frac, levels, dpde_l,
                                 chg_leak_lt, chg_open_lt)
    else:
        raise StisPixCteError('Invalid detector: StisPixCteCorr only supports STIS CCD.')

    return corrected


def AddYCte(infile, outfile, shift_nit=None, units=None):
    """
    Add CTE blurring to input image using an inversion of the CTE correction
    code.

    .. note::

        No changes are made to the error or data quality arrays.

        Data should not have bias or prescan regions.

        Image must have PCTETAB, DETECTOR, TEXPSTRT, ATODGAIN,
        and NEXTEND header keywords.

    Parameters
    ----------
    infile : str
        Filename of image to be blurred. Should have the PCTETAB header
        keyword pointing to the PCTETAB reference file.

    outfile : str
        Filename of blurred output image.

    shift_nit : int, optional
        Number of times column is shifted during simulated readout.
        If None, takes value from PCTETAB header SHFT_NIT keyword.

    units : {None,'electrons','counts'}, optional
        If 'electrons', the input image is assumed to have units of electrons
        and no gain operations are performed.
        If 'counts', the data are assumed to be in DN and they are converted
        to electrons before CTE blurring is performed. The ATODGN* keywords
        from the primary header are used for the conversions.
        If None, the BUNIT keyword from the science extension headers is used
        to set the unit behavior.
        Defaults to None.
        For STIS data, this argument is ignored; the units are assumed
        to be counts, so the data will be converted to electrons.

    Raises
    ------
    ValueError
        If the number of extensions is not a multiple of three.

    stistools.StisPixCteCorr.StisPixCteError
        If the input image comes from an imcompatible detector.

    """

    # copy infile to outfile
    shutil.copyfile(infile, outfile)

    # open file for blurring
    pf_out = pyfits.open(outfile, mode='update')

    # For checking that the detector is supported
    detector = pf_out['PRIMARY'].header['DETECTOR']

    atodgain = pf_out['PRIMARY'].header.get('ATODGAIN', default=1.)

    # For epoch-specific operations
    expstart = pf_out['PRIMARY'].header['TEXPSTRT']
    
    nextend = pf_out['PRIMARY'].header.get('NEXTEND', default=EXTN_PER_IMSET)
    nimsets = nextend // EXTN_PER_IMSET
    if nextend != nimsets * EXTN_PER_IMSET:
        raise ValueError('Number of extensions is %d; must be a multiple'
                         ' of %d.' % (nextend, EXTN_PER_IMSET))

    # This is just for CCD.
    if detector != 'CCD':
        os.remove(outfile)
        raise StisPixCteError('Invalid detector: StisPixCteCorr only'
                              ' supports STIS CCD.')

    # Read CTE params from file
    pctefile = pf_out['PRIMARY'].header['PCTETAB']
    pardict = _PixCteParams(pctefile, expstart)

    cte_frac = pardict['cte_frac']
    q_dtde = pardict['q_dtde']
    dtde_l = pardict['dtde_l']
    psi_node = pardict['psi_node']
    chg_leak = pardict['chg_leak']
    levels = pardict['levels']
    col_scale = pardict['col_scale']

    if shift_nit is None:
        shft_nit = pardict['shft_nit']
    else:
        shft_nit = shift_nit

    # The following are only used to update the header of the output, blurred image
    sim_nit = pardict['sim_nit']
    rn_clip = pardict['read_noise']

    # N in charge tail
    chg_leak_kt, chg_open_kt = pcfy.InterpolatePsi(chg_leak, psi_node)
    del chg_leak, psi_node

    # dtde_q: Marginal PHI at a given chg level.
    # q_pix_array: Maps Q (cumulative charge) to P (dependent var).
    # pix_q_array: Maps P to Q.
    dtde_q = pcfy.InterpolatePhi(dtde_l, q_dtde, shft_nit)
    del dtde_l, q_dtde

    # finish interpolation along the Q dimension and reduce arrays to contain
    # only info at the levels specified in the levels array
    chg_leak_lt, chg_open_lt, dpde_l = \
      pcfy.FillLevelArrays(chg_leak_kt, chg_open_kt, dtde_q, levels)
    del chg_leak_kt, chg_open_kt, dtde_q
    
    ccdamp = pf_out['PRIMARY'].header.get('CCDAMP', default='D')
    if ccdamp == 'A' or ccdamp == 'B':
        readout_dir = 1
    elif ccdamp == 'C' or ccdamp == 'D':
        readout_dir = -1
    else:
        raise ValueError("Amplifier %s is unrecognized!" % ccdamp)
    
    ########################################
    # perform CTE blurring for each image set
    ########################################
    for imset in range(nimsets):
        # get data
        extn = 1 + imset * EXTN_PER_IMSET
        # [::-1,::-1] flips the data so that pixel [0, 0] is closest to
        # the readout, since that's what the algorithm expects
        # (this assumes STIS amp D was used)
        # Modified (SL):  Does readout_dir==+1 work for 'A' and 'B'?
        scidata = pf_out[extn].data[::readout_dir,::readout_dir].copy().astype(numpy.float)
        # convert to electrons
        scidata *= atodgain

        # set up cte frac array; flip amp D info from col_scale
        col_scale_temp = col_scale[ccdamp][:scidata.shape[1]][::readout_dir]
        cte_frac_arr = cte_frac * numpy.ones_like(scidata) * col_scale_temp
        cte_frac_arr = cte_frac_arr * \
            numpy.arange(1,scidata.shape[0]+1).reshape((scidata.shape[0],1))
        cte_frac_arr /= 1024.

        # call CTE blurring routine. data must be in units of electrons.
        print 'Performing CTE blurring for science extension %d.' % extn

        t1 = time.time()
        cordata = _AddYCte(detector, scidata, cte_frac_arr, shft_nit,
                           levels, dpde_l, chg_leak_lt, chg_open_lt)
        t2 = time.time()

        print 'AddYCte took {:.3f} seconds for science extension {:d}.'.format(t2-t1, extn)

        # convert blurred data back to DN
        cordata /= atodgain

        # copy blurred data back to image, flipping to its original orientation
        pf_out[extn].data[:,:] = cordata.astype(numpy.float32)[::readout_dir,::readout_dir]
        # end for imset in range(nimsets)

    # Update header
    pf_out['PRIMARY'].header['PCTEFRAC'] = cte_frac
    pf_out['PRIMARY'].header['PCTERNCL'] = rn_clip
    pf_out['PRIMARY'].header['PCTESMIT'] = sim_nit
    pf_out['PRIMARY'].header['PCTESHFT'] = shft_nit
    pf_out['PRIMARY'].header.add_history('CTE blurring performed by PixCteCorr.AddYCte')

    # close image
    pf_out.close()


def _AddYCte(detector, input_data, cte_frac, shft_nit, levels, dpde_l,
             chg_leak_lt, chg_open_lt):
    """
    Apply STIS CTE blurring to input data.

    Parameters
    ----------
    detector : str
        DETECTOR from header.
        Currently only 'CCD' is supported.

    input_data : ndarray
        Data in need of CTE correction. For proper results cte_data[0,x] should
        be next to the readout register and cte_data[-1,x] should be furthest.
        Data are processed a column at a time, e.g. cte_data[:,x] is corrected,
        then cte_data[:,x+1] and so on.

    cte_frac : ndarray
        CTE scaling image combining the time dependent and column-by-column
        CTE factors. Should be same shape as `cte_data`.

    shft_nit : int
        Number of readout shifts to do.

    levels : ndarray
        Levels at which CTE is evaluated as read from PCTEFILE.

    dpde_l : ndarray
        Parameterized amount of charge in CTE trails as a function of
        specific charge levels, as returned by _FillLevelArrays.

    chg_leak_lt : ndarray
        Tail profile data at charge levels specified by levels, as returned
        by _FillLevelArrays.

    chg_open_lt : ndarray
        Tail profile data at charge levels specified by levels, as returned
        by _FillLevelArrays.

    Returns
    -------
    blurred : ndarray
        Data CTE correction algorithm applied.
        Same size and shape as input_data.

    """

    if detector == 'CCD':
        blurred = pcfy.AddYCte(input_data, shft_nit, cte_frac,
                                levels, dpde_l, chg_leak_lt, chg_open_lt)
    else:
        raise StisPixCteError('Invalid detector: StisPixCteCorr only supports STIS CCD.')

    return blurred


#--------------------------
# TEAL Interface functions
#--------------------------
def run(configObj):
    """
    TEAL interface for the `CteCorr` function.

    """
    CteCorr(configObj['inFits'],
            outFits=configObj['outFits'],
            read_noise=configObj['read_noise'],
            noise_model=configObj['noise_model'],
            oversub_thresh=configObj['oversub_thresh'],
            sim_nit=configObj['sim_nit'],
            shift_nit=configObj['shift_nit'])


def getHelpAsString():
    """
    Returns documentation on the `CteCorr` function. Required by TEAL.

    """
    return CteCorr.__doc__
